/*
 * vzct_driver.c: core driver functions for managing
 * Virtuozzo Containers
 *
 * Copyright (C) 2017 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>
#include <vzctl/libvzctl.h>
#include <vzctl/vzerror.h>
#include <linux/netlink.h>
#include <gio/gio.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>

#include "virlog.h"
#include "viraccessapicheck.h"
#include "viraccessapicheckext.h"
#include "viralloc.h"
#include "virdomainobjlist.h"
#include "vireventthread.h"
#include "virutil.h"
#include "datatypes.h"
#include "virfile.h"
#include "virstring.h"
#include "virthreadpool.h"
#include "configmake.h"
#include "vircommand.h"
#include "virportallocator.h"
#include "virpidfile.h"
#include "virdomainsnapshotobjlist.h"
#include "snapshot_conf.h"
#include "virtime.h"
#include "virthreadjob.h"
#include "domain_event.h"
#include "virprocess.h"
#include "virfdstream.h"

#include "vzct_driver.h"

#define VIR_FROM_THIS VIR_FROM_VZCT

VIR_LOG_INIT("vzct.vzct_driver");

#define VZCT_ERROR_MESSAGE_CODE(func, code) \
do { \
    const char *check_err = vzctl2_get_last_error(); \
    if (check_err && *check_err) { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed with code %d: %s"), \
                       #func,  code, check_err); \
    } else { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed with code %d"), \
                       #func,  code); \
    } \
} while (0)

#define VZCT_CHECK(label, func) do { \
    int check_rc; \
    if ((check_rc = func)) { \
        VZCT_ERROR_MESSAGE_CODE(func, check_rc); \
        goto label; \
    } \
} while (0)

#define VZCT_CHECK_RET(ret, func) do { \
    int check_rc; \
    if ((check_rc = func)) { \
        VZCT_ERROR_MESSAGE_CODE(func, check_rc); \
        return ret; \
    } \
} while (0)

#define VZCT_ERROR_MESSAGE(func) \
do { \
    const char *check_err = vzctl2_get_last_error(); \
    if (check_err && *check_err) { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed: %s"), \
                       #func,  check_err); \
    } else { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed"), \
                       #func); \
    } \
} while (0)

#define VZCT_CHECK_PTR(label, ptr, func) do { \
    if (!(ptr = func)) { \
        VZCT_ERROR_MESSAGE(func); \
        goto label; \
    } \
} while (0)

#define VZCT_CHECK_PTR_RET(ret, ptr, func) do { \
    if (!(ptr = func)) { \
        VZCT_ERROR_MESSAGE(func); \
        return ret; \
    } \
} while (0)

#define VZCT_CHECK_LOCKED(label, env, status, func) do { \
    int check_rc; \
    int lockfd; \
    if ((lockfd = vzctl2_env_lock(env, status)) < 0) { \
        VZCT_ERROR_MESSAGE(vzctl2_env_lock); \
        goto label; \
    } \
    if ((check_rc = func)) { \
        VZCT_ERROR_MESSAGE_CODE(func, check_rc); \
    } \
    vzctl2_env_unlock(env, lockfd); \
    if (check_rc) \
        goto label; \
} while (0)

#define VZCT_PRIVATE_DIR ".libvirt"
#define VZCT_SNAPHOTS_DIR VZCT_PRIVATE_DIR"/snapshots"

#define VZCT_NET_VENET0 "venet0"

#define VZCT_DISK_ROOT_UUID "{00000000-0000-0000-0000-000000000000}"

VIR_ENUM_DECL(vzctlEvent);
VIR_ENUM_IMPL(vzctlEvent,
              VZCTL_ENV_UMOUNT + 1,
              "unknown",
              "created",
              "registered",
              "unregistered",
              "deleted",
              "config_changed",
              "started",
              "stopped",
              "suspended",
              "classes_config_changed",
              "shaping_config_changed",
              "net_config_changed",
              "umount",
);


typedef struct _vzctDriverConfig vzctDriverConfig;
typedef vzctDriverConfig *vzctDriverConfigPtr;
struct _vzctDriverConfig {
    char *stateDir;
};

typedef struct _vzctDriver vzctDriver;
typedef struct _vzctDriver *vzctDriverPtr;
struct _vzctDriver {
    virMutex lock;
    virDomainXMLOptionPtr xmlopt;
    vzctDriverConfigPtr config;
    virCapsPtr caps;
    virEventThread *eventThread;

    virDomainObjListPtr domains;

    vzevt_handle_t *eventdHandle;
    GSocket *eventdSocket;
    GSource *eventdSource;

    int netlinkFd;
    GSocket *netlinkSocket;
    GSource *netlinkSource;

    virPortAllocatorRangePtr vncPorts;

    virThreadPoolPtr workers;
    virObjectEventStatePtr eventsManager;

    int lastconnid;
};

typedef enum {
    VZCT_DOMAIN_JOB_NONE = 0,
    VZCT_DOMAIN_JOB_QUERY = 1,
    VZCT_DOMAIN_JOB_MODIFY = 2,
    VZCT_DOMAIN_JOB_MIGRATION = 3,
} vzctDomainJobType;

typedef struct _vzctDomainJobStats vzctDomainJobStats;
typedef vzctDomainJobStats *vzctDomainJobStatsPtr;
struct _vzctDomainJobStats {
    unsigned long long timeElapsed;
};

typedef struct _vzctDomainExecProcess vzctDomainExecProcess;
typedef vzctDomainExecProcess *vzctDomainExecProcessPtr;
struct _vzctDomainExecProcess {
    struct vzctl_exec_handle *handle;
    int conn_id;
    bool orphan;
    bool terminated;
};

typedef struct _vzctDomainObjPrivate vzctDomainObjPrivate;
typedef vzctDomainObjPrivate *vzctDomainObjPrivatePtr;
struct _vzctDomainObjPrivate {
    virDomainSnapshotObjListPtr snapshots;
    char *bundlePath;
    pid_t vncPid;

    virCond jobCond;
    vzctDomainJobType job;
    const char *jobAPI;
    vzctDomainJobType sideJob;
    const char *sideJobAPI;
    bool allowSideJob;
    int mig_conn_id;
    unsigned long long jobStarted;
    pid_t migrationPid;
    bool abortJob;

    vzctDomainExecProcessPtr *exec_procs;
    size_t exec_nprocs;
    size_t exec_nprocs_max;
    int cleanupTimer;
};

#define VZCT_DOMAIN_DISK_UUID_SIZE 39

typedef struct _vzctDomainDiskPrivate vzctDomainDiskPrivate;
typedef vzctDomainDiskPrivate *vzctDomainDiskPrivatePtr;
struct _vzctDomainDiskPrivate {
    virObject parent;

    /* Size is copied from libvzctl header. Looks like common uuid format but
     * with braces. */
    char uuid[VZCT_DOMAIN_DISK_UUID_SIZE];
};

typedef enum {
    VZCT_DRIVER_JOB_EVENT = 0,
    VZCT_DRIVER_JOB_CONNECTION_CLOSED = 1,
    VZCT_DRIVER_JOB_CLEANUP = 2,
} vzctDriverJobType;

typedef enum {
    VZCT_DOMAIN_EVENT_NONE = 0,
    VZCT_DOMAIN_EVENT_STARTED,
    VZCT_DOMAIN_EVENT_STOPPED,
    VZCT_DOMAIN_EVENT_DEFINED,
    VZCT_DOMAIN_EVENT_UNDEFINED,
    VZCT_DOMAIN_EVENT_UPDATED,
} vzctDomainEventType;

typedef struct _vzctDriverJob vzctDriverJob;
typedef vzctDriverJob *vzctDriverJobPtr;
struct _vzctDriverJob {
    vzctDriverJobType type;
    vzctDomainEventType eventType;
    char *ctid;
    int conn_id;
    virDomainObjPtr dom;
};


static virClassPtr vzctDomainDiskPrivateClass;
static void vzctDomainDiskPrivateDispose(void *obj);

static int
vzctDomainDiskPrivateOnceInit(void)
{
    if (!VIR_CLASS_NEW(vzctDomainDiskPrivate, virClassForObject()))
        return -1;

    return 0;
}

VIR_ONCE_GLOBAL_INIT(vzctDomainDiskPrivate);

static vzctDriverPtr vzct_driver;

static void
vzctDriverConnectClose(vzctDriverPtr driver);

static virDomainObjPtr
vzctAddDomain(vzctDriverPtr driver,
              const char *id,
              virDomainObjPtr dom);

static int
vzctUpdateDomain(vzctDriverPtr driver,
                 virDomainObjPtr dom);

static int
vzctDomainStartFull(vzctDriverPtr driver,
                    virDomainObjPtr dom);

static void
vzctDomainCleanupTimeout(int timer, void *opaque);

G_DEFINE_AUTOPTR_CLEANUP_FUNC(virDomain, virObjectUnref);

static void
vzctDomainClearVolumes(virDomainDefPtr def,
                       vzctl_env_handle *env);

static virObjectPtr
vzctDomainDiskPrivateNew(void)
{
    vzctDomainDiskPrivatePtr priv;

    if (vzctDomainDiskPrivateInitialize() < 0)
        return NULL;

    if (!(priv = virObjectNew(vzctDomainDiskPrivateClass)))
        return NULL;

    return (virObjectPtr) priv;
}


static void
vzctDomainDiskPrivateDispose(void *obj G_GNUC_UNUSED)
{
}

static void
vzctDomainFinishSnapshots(virDomainObjPtr dom);

static int
vzctDomainReadSnapshots(virDomainObjPtr dom);

static int
vzctDomainLoadSnapshotMetadata(vzctDriverPtr driver,
                               virDomainObjPtr dom);


static void
vzctDriverJobFree(vzctDriverJobPtr job)
{
    if (!job)
        return;

    VIR_FREE(job->ctid);
    VIR_FREE(job);
}


static void
vzctDomainStop(vzctDriverPtr driver,
               virDomainObjPtr dom);

static bool
vzctMergeConfigs(virDomainDefPtr def,
                 virDomainDefPtr drvDef);

static virDomainDefPtr
vzctLoadDomainConfig(vzctDriverPtr driver,
                     const char *ctid,
                     char **bundlePath);


/*
 * We need this function to copy disk private.
 */
static virDomainDefPtr
vzctDomainDefCopy(virDomainDefPtr def,
                  virDomainXMLOptionPtr xmlopt,
                  bool migratable)
{
    unsigned int format_flags = VIR_DOMAIN_DEF_FORMAT_SECURE |
                                VIR_DOMAIN_DEF_FORMAT_DISK_PRIVATE;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                               VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                                VIR_DOMAIN_DEF_PARSE_DISK_PRIVATE;
    g_autofree char *xml = NULL;
    virBuffer buf = VIR_BUFFER_INITIALIZER;

    if (migratable)
        format_flags |= VIR_DOMAIN_DEF_FORMAT_INACTIVE |
                        VIR_DOMAIN_DEF_FORMAT_MIGRATABLE;

    if (virDomainDefFormatInternal(def, xmlopt, &buf, format_flags) < 0 ||
        !(xml = virBufferContentAndReset(&buf)))
        return NULL;

    return virDomainDefParseString(xml, xmlopt, NULL, parse_flags);
}


static bool
vzctDomainJobAllowed(vzctDomainObjPrivatePtr priv,
                     vzctDomainJobType job)
{
    if (!priv->job)
        return true;

    return priv->allowSideJob &&
           !priv->sideJob &&
           job == VZCT_DOMAIN_JOB_QUERY;
}


/*
 * Acquire job.
 * @timeout: -1 - wait indefinetly
 *            0 - don't wait
 *           >0 - wait for specified timeout in milliseconds
 */
static int
vzctDomainBeginJobInternal(virDomainObjPtr dom,
                           vzctDomainJobType job,
                           long long timeout)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    unsigned long long then = 0;

    if (timeout > 0) {
        unsigned long long now;

        if (virTimeMillisNow(&now) < 0)
            return -1;

        then = now + timeout;
    }

    while (!vzctDomainJobAllowed(priv, job)) {
        int rc;

        if (timeout == 0)
            return -1;

        if (timeout > 0)
            rc = virObjectWaitUntil(dom, &priv->jobCond, then);
        else
            rc = virObjectWait(dom, &priv->jobCond);

        if (rc == 0)
            continue;

        if (errno == ETIMEDOUT) {
            if (priv->job && priv->sideJob) {
                virReportError(VIR_ERR_OPERATION_TIMEOUT,
                               _("cannot acquire lock for %s"
                                 " (held by job=%s sideJob=%s)"),
                               virThreadJobGet(),
                               priv->jobAPI, priv->sideJobAPI);
            } else {
                virReportError(VIR_ERR_OPERATION_TIMEOUT,
                               _("cannot acquire lock for %s"
                                 " (held by job=%s)"),
                               virThreadJobGet(), priv->jobAPI);
            }
        } else {
            virReportSystemError(errno, _("cannot acquire lock for %s"),
                                 virThreadJobGet());
        }

        return -1;
    }

    if (dom->removing) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];

        virUUIDFormat(dom->def->uuid, uuidstr);
        if (timeout > 0)
            virReportError(VIR_ERR_NO_DOMAIN,
                           _("no domain name=%s uuid=%s ctid=%s"),
                           dom->def->name, uuidstr, dom->def->extraId);
        else
            VIR_DEBUG("no domain name=%s uuid=%s ctid=%s",
                      dom->def->name, uuidstr, dom->def->extraId);
        return -1;
    }

    if (job == VZCT_DOMAIN_JOB_MIGRATION &&
        virTimeMillisNow(&priv->jobStarted) < 0)
        return -1;

    if (priv->job) {
        VIR_DEBUG("begin side job %s", virThreadJobGet());
        priv->sideJob = job;
        priv->sideJobAPI = virThreadJobGet();
    } else {
        VIR_DEBUG("begin main job %s", virThreadJobGet());
        priv->job = job;
        priv->jobAPI = virThreadJobGet();
    }

    return 0;
}

#define VZCT_JOB_WAIT_TIME (30 * 1000)

static int
vzctDomainBeginJob(virDomainObjPtr dom,
                   vzctDomainJobType job)
{
    return vzctDomainBeginJobInternal(dom, job, VZCT_JOB_WAIT_TIME);
}


static void
vzctDomainAllowSideJob(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    priv->allowSideJob = true;
}


static void
vzctDomainDisableSideJob(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;

    /* Disable here so no new side job can start */
    priv->allowSideJob = false;
    while (priv->sideJob) {
        if (virObjectWait(dom, &priv->jobCond) < 0) {
            VIR_WARN("cannot wait side job finishing, errno=%d", errno);
            return;
        }
    }
}


static void
vzctDomainEndJob(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;

    if (priv->sideJob) {
        VIR_DEBUG("end side job %s", priv->sideJobAPI);
        priv->sideJob = VZCT_DOMAIN_JOB_NONE;
        priv->sideJobAPI = NULL;
    } else {
        VIR_DEBUG("end main job %s", priv->jobAPI);
        priv->job = VZCT_DOMAIN_JOB_NONE;
        priv->jobAPI = NULL;
        priv->mig_conn_id = 0;
        priv->jobStarted = 0;
        priv->abortJob = false;
        if (priv->allowSideJob)
            VIR_WARN("unexpected job end");
        priv->allowSideJob = false;
    }

    virCondBroadcast(&priv->jobCond);
}


static void
vzctDomainDropJob(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;

    VIR_DEBUG("drop job %s", priv->jobAPI);
    if (priv->sideJob) {
        VIR_DEBUG("convert side job to main job for %s", priv->sideJobAPI);
        priv->job = priv->sideJob;
        priv->jobAPI = priv->sideJobAPI;
        priv->sideJob = VZCT_DOMAIN_JOB_NONE;
        priv->sideJobAPI = NULL;
    } else {
        priv->job = VZCT_DOMAIN_JOB_NONE;
        priv->jobAPI = NULL;
        virCondBroadcast(&priv->jobCond);
    }

    priv->allowSideJob = false;
    priv->mig_conn_id = 0;
    priv->jobStarted = 0;
    priv->abortJob = false;
}


static gboolean
vzctNetlinkEvent(GSocket *socket G_GNUC_UNUSED,
                 GIOCondition cond,
                 gpointer opaque)
{
    vzctDriverPtr driver = opaque;
    struct sockaddr_nl addr;
    struct msghdr msg = { .msg_control = 0 };
    struct iovec iov;
    char buf[512];
    char *event, *ctid, *saveptr;
    vzctDriverJobPtr job = NULL;
    int len;

    virMutexLock(&driver->lock);

    if (!driver->netlinkSource)
        goto cleanup;

    if (cond & ~G_IO_IN) {
        if (cond & G_IO_OUT)
            VIR_WARN("Got writable event on vz event descriptor.");
        else if (cond & G_IO_HUP)
            VIR_DEBUG("Got HANGUP on vz event descriptor.");
        else if (cond & G_IO_ERR)
            VIR_WARN("Got ERROR on vz event descriptor.");

        /* Generally speaking we should not close connection as we go
         * out of sync. But these events should never happend on netlink
         * so this is just defencing from event loop eating 100% CPU. */
        vzctDriverConnectClose(driver);
        goto cleanup;
    }

    /* check for spurious */
    if (!(cond & G_IO_IN))
        goto cleanup;

    msg.msg_name = (void*)&addr;
    msg.msg_namelen = sizeof(addr);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    iov.iov_base = buf;
    iov.iov_len = sizeof(buf) - 1;

    while ((len = recvmsg(driver->netlinkFd, &msg, MSG_DONTWAIT)) < 0 &&
           errno == EINTR)
        ;

    if (len < 0) {
        if (errno != EAGAIN)
            VIR_ERROR(_("Can not read event message, error: %d"), errno);
        goto cleanup;
    }

    if (!len)
        goto cleanup;

    buf[len] = '\0';

    if (!(event = strtok_r(buf, "@", &saveptr))) {
        VIR_ERROR(_("Can not parse event type"));
        goto cleanup;
    }

    if (!(ctid = strtok_r(NULL, "@", &saveptr))) {
        VIR_ERROR(_("Can not parse container id"));
        goto cleanup;
    }

    VIR_LOG_VZ("received netlink event ctid=%s event=%s", ctid, event);

    /* Case for changing state in worker thread. During CT start
     * thru vzct driver first netlink event is arrived and blocked
     * on acquiring domain lock so the entire event loop is blocked.
     * Next in the process of starting domain prlsrvctl is called
     * to resolve network to bridge I guess. prlsrvctl in turn query
     * libvirtd and the latter is unresponsive as event loop is blocked */
    VIR_ALLOC(job);
    job->type = VZCT_DRIVER_JOB_EVENT;
    job->ctid = g_strdup(ctid);
    if (STREQ(event, "ve-start")) {
        job->eventType = VZCT_DOMAIN_EVENT_STARTED;
    } else if (STREQ(event, "ve-stop")) {
        job->eventType = VZCT_DOMAIN_EVENT_STOPPED;
    } else {
        VIR_DEBUG("Unexpected event type %s", event);
        goto cleanup;
    }

    if (virThreadPoolSendJob(driver->workers, 0, job) < 0)
        goto cleanup;
    job = NULL;

 cleanup:
    vzctDriverJobFree(job);
    virMutexUnlock(&driver->lock);
    return G_SOURCE_CONTINUE;
}


static gboolean
vzctEventdEvent(GSocket *socket G_GNUC_UNUSED,
                GIOCondition cond,
                gpointer opaque)
{
    vzctDriverPtr driver = opaque;
    struct vzctl_state_evt vzevent;
    vzctDriverJobPtr job = NULL;

    virMutexLock(&driver->lock);

    if (!driver->eventdSource)
        goto cleanup;

    if (cond & ~G_IO_IN) {
        if (cond & G_IO_OUT)
            VIR_WARN("Got writable event on vz event descriptor.");
        else if (cond & G_IO_HUP)
            VIR_DEBUG("Got HANGUP on vz event descriptor.");
        else if (cond & G_IO_ERR)
            VIR_WARN("Got ERROR on vz event descriptor.");

        /* Generally speaking we should not close connection as we go
         * out of sync. But these events should never happend on netlink
         * so this is just defencing from event loop eating 100% CPU. */
        vzctDriverConnectClose(driver);
        goto cleanup;
    }

    /* check for spurious */
    if (!(cond & G_IO_IN))
        goto cleanup;

    /* FIXME Unfortunately this API returns -1 in case of no event and in
     * case of error thus we can miss some errors now. But errors are
     * OOM only so they are unlikely. Treat as no event. */
    if (vzctl2_get_state_evt(driver->eventdHandle,
                             &vzevent, sizeof(vzevent)) < 0)
        goto cleanup;

    VIR_LOG_VZ("received eventd event ctid=%s event=%s", vzevent.ctid,
               NULLSTR(vzctlEventTypeToString(vzevent.state)));

    VIR_ALLOC(job);
    job->type = VZCT_DRIVER_JOB_EVENT;
    job->ctid = g_strdup(vzevent.ctid);
    switch (vzevent.state) {
    case VZCTL_ENV_CREATED:
    case VZCTL_ENV_REGISTERED:
        job->eventType = VZCT_DOMAIN_EVENT_DEFINED;
        break;
    case VZCTL_ENV_UNREGISTERED:
    case VZCTL_ENV_DELETED:
        job->eventType = VZCT_DOMAIN_EVENT_UNDEFINED;
        break;
    case VZCTL_ENV_CONFIG_CHANGED:
    case VZCTL_ENV_NET_CONFIG_CHANGED:
        job->eventType = VZCT_DOMAIN_EVENT_UPDATED;
        break;
    default:
        goto cleanup;
    }

    if (virThreadPoolSendJob(driver->workers, 0, job) < 0)
        goto cleanup;
    job = NULL;

 cleanup:
    vzctDriverJobFree(job);
    virMutexUnlock(&driver->lock);
    return G_SOURCE_CONTINUE;
}


static void
vzctDriverConnectClose(vzctDriverPtr driver)
{
    VIR_DEBUG("Closing connection to vzeventd.");

    if (driver->eventdSource) {
        g_source_destroy(driver->eventdSource);
        vir_g_source_unref(driver->eventdSource,
                           virEventThreadGetContext(driver->eventThread));
        driver->eventdSource = NULL;
        g_object_unref(driver->eventdSocket);
        driver->eventdSocket = NULL;
    }
    vzctl2_unregister_evt(driver->eventdHandle);
    driver->eventdHandle = NULL;

    if (driver->netlinkSource) {
        g_source_destroy(driver->netlinkSource);
        vir_g_source_unref(driver->netlinkSource,
                           virEventThreadGetContext(driver->eventThread));
        driver->netlinkSource = NULL;
        g_object_unref(driver->netlinkSocket);
        driver->netlinkSocket = NULL;
        driver->netlinkFd = -1;
    }
}


static char *
vzctStateFile(vzctDriverConfigPtr config, const unsigned char *uuid)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    char *path;

    virUUIDFormat(uuid, uuidstr);
    path = g_strdup_printf("%s/%s.xml", config->stateDir, uuidstr);
    return path;
}


static char *
vzctMigFile(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;

    return g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/status.migrating.xml",
                           priv->bundlePath);
}


static char *
vzctVncPidFile(vzctDriverConfigPtr config, const unsigned char *uuid)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    char *path;

    virUUIDFormat(uuid, uuidstr);
    path = g_strdup_printf("%s/%s.vnc.pid", config->stateDir, uuidstr);
    return path;
}


static int
vzctDomainSaveConfig(vzctDriverPtr driver, virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    virDomainDefPtr def = virDomainObjIsActive(dom) ? dom->newDef : dom->def;
    g_autofree char *path = NULL;

    path = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/config.xml", priv->bundlePath);

    return virDomainDefSavePath(def, driver->xmlopt, path,
                                VIR_DOMAIN_DEF_FORMAT_DISK_PRIVATE);
}


static int
vzctDomainStateLoad(vzctDriverPtr driver,
                    virDomainObjPtr dom,
                    const unsigned char *uuid)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    g_autoptr(virDomainObj) saved = NULL;
    g_autofree char *statefile = vzctStateFile(driver->config, uuid);
    g_autofree char *pidfile = vzctVncPidFile(driver->config, uuid);
    unsigned int parseFlags = VIR_DOMAIN_DEF_PARSE_STATUS |
                              VIR_DOMAIN_DEF_PARSE_ACTUAL_NET |
                              VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                              VIR_DOMAIN_DEF_PARSE_ALLOW_POST_PARSE_FAIL;

    if (!(saved = virDomainObjParseFile(statefile, driver->xmlopt, parseFlags)))
        return -1;
    virObjectUnlock(saved);

    if (dom->def) {
        virDomainDefPtr def;

        /* FIXME Use active config from vzctl when it will be available. */
        if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
            return -1;
        dom->newDef = g_steal_pointer(&dom->def);
        dom->def = g_steal_pointer(&def);
        if (vzctMergeConfigs(dom->def, saved->def) &&
            virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            return -1;
    } else {
        dom->def = g_steal_pointer(&saved->def);
    }

    virDomainObjSetState(dom, VIR_DOMAIN_RUNNING, -1);
    dom->def->id = 0;

    if (dom->def->ngraphics) {
        virDomainGraphicsDefPtr gr = dom->def->graphics[0];

        gr->data.vnc.portReserved = true;
        virPortAllocatorSetUsed(gr->data.vnc.port);

        if (virFileExists(pidfile) &&
            virPidFileReadPath(pidfile, &priv->vncPid) < 0)
            return -1;
    }

    /*
     * Timer can be already initialized if we are recovering from unsynced
     * state.
     */
    if (!priv->cleanupTimer) {
        virObjectRef(dom);
        if ((priv->cleanupTimer = virEventAddTimeout(-1,
                                                     vzctDomainCleanupTimeout,
                                                     dom,
                                                     virObjectFreeCallback)) < 0) {
            virObjectUnref(dom);
            return -1;
        }
    }

    return 0;
}


static void
vzctCleanupDriverConfigs(vzctDriverPtr driver)
{
    DIR *dir;
    struct dirent *entry;
    g_autofree char *path = NULL;
    const char *stateDir = driver->config->stateDir;

    if (virDirOpenIfExists(&dir, stateDir) <= 0)
        return;

    while (virDirRead(dir, &entry, stateDir) > 0) {
        unsigned char uuid[VIR_UUID_BUFLEN];
        char *uuidstr = entry->d_name;
        virDomainObjPtr dom;

        /* Domain states are stored with filenames uuid.xml */
        if (!virStringStripSuffix(uuidstr, ".xml") ||
            virUUIDParse(uuidstr, uuid) < 0)
            continue;

        if ((dom = virDomainObjListFindByUUID(driver->domains, uuid))) {
            virDomainObjEndAPI(&dom);
            continue;
        }

        if (!(dom = virDomainObjNew(driver->xmlopt)))
            continue;

        if (vzctDomainStateLoad(driver, dom, uuid) < 0) {
            virObjectUnref(dom);
            continue;
        }

        vzctDomainStop(driver, dom);
        virObjectUnref(dom);
    }
    virDirClose(&dir);
}


static int
vzctDriverConnectOpen(vzctDriverPtr driver)
{
    struct sockaddr_nl addr = { 0 };
    struct vzctl_ids *ids = NULL;
    g_autoptr(GError) gerr = NULL;
    int fd;
    int ndoms;
    size_t i;
    int ret = -1;

    virMutexLock(&driver->lock);

    VZCT_CHECK(cleanup, vzctl2_init_log("libvirtd"));
    VZCT_CHECK(cleanup, vzctl2_set_log_quiet(1));
    VZCT_CHECK(cleanup, vzctl2_lib_init());

    if (!(driver->eventThread = virEventThreadNew("vzct")))
        goto cleanup;

    VZCT_CHECK(cleanup,
               vzctl2_register_evt(&driver->eventdHandle));

    driver->netlinkFd = -1;

    fd = vzctl2_get_evt_fd(driver->eventdHandle);
    if (!(driver->eventdSocket = g_socket_new_from_fd(fd, &gerr))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unable to create socket object: %s"), gerr->message);
        goto cleanup;
    }

    driver->eventdSource = g_socket_create_source(driver->eventdSocket,
                                                  G_IO_IN, NULL);
    g_source_set_callback(driver->eventdSource,
                          (GSourceFunc)vzctEventdEvent,
                          driver, NULL);
    g_source_attach(driver->eventdSource,
                    virEventThreadGetContext(driver->eventThread));

    if ((driver->netlinkFd = socket(PF_NETLINK, SOCK_DGRAM, 31)) < 0) {
        virReportSystemError(errno,
                             "%s", _("failed to netlink socket"));
        goto cleanup;
    }

    if (virSetCloseExec(driver->netlinkFd) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to set netlink socket close-on-exec flag"));
        goto cleanup;
    }

    addr.nl_family = AF_NETLINK;
    addr.nl_groups = 0x01; /* multicast */

    if (bind(driver->netlinkFd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to bind netlink socket"));
        goto cleanup;
    }

    if (!(driver->netlinkSocket = g_socket_new_from_fd(driver->netlinkFd, &gerr))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unable to create socket object: %s"), gerr->message);
        goto cleanup;
    }

    driver->netlinkSource = g_socket_create_source(driver->netlinkSocket,
                                                   G_IO_IN, NULL);
    g_source_set_callback(driver->netlinkSource,
                          (GSourceFunc)vzctNetlinkEvent,
                          driver, NULL);
    g_source_attach(driver->netlinkSource,
                    virEventThreadGetContext(driver->eventThread));

    VZCT_CHECK_PTR(cleanup, ids, vzctl2_alloc_env_ids());

    if ((ndoms = vzctl2_get_env_ids_by_state(ids, ENV_STATUS_EXISTS)) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("vzctl2_get_env_ids_by_state failed"));
        goto cleanup;
    }

    if (!(driver->domains = virDomainObjListNewFlags(true)))
        goto cleanup;

    for (i = 0; i < ndoms; i++) {
        virDomainObjPtr dom = vzctAddDomain(driver, ids->ids[i], NULL);
        virDomainObjEndAPI(&dom);
    }

    ret = 0;

 cleanup:
    virMutexUnlock(&driver->lock);
    if (ids)
        vzctl2_free_env_ids(ids);
    return ret;
}


static virDrvOpenStatus
vzctConnectOpen(virConnectPtr conn,
                virConnectAuthPtr auth G_GNUC_UNUSED,
                virConfPtr conf G_GNUC_UNUSED,
                unsigned int flags)
{
    virCheckFlags(VIR_CONNECT_RO, VIR_DRV_OPEN_ERROR);

    if (!conn->uri)
        return VIR_DRV_OPEN_DECLINED;

    if (!conn->uri->scheme)
        return VIR_DRV_OPEN_DECLINED;

    if (STRNEQ(conn->uri->scheme, "vzct"))
        return VIR_DRV_OPEN_DECLINED;

    /* Remote driver should handle these. */
    if (conn->uri->server)
        return VIR_DRV_OPEN_DECLINED;

    /* From this point on, the connection is for us. */
    if (STRNEQ_NULLABLE(conn->uri->path, "/system")) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unexpected Virtuozzo URI path '%s', try vzct:///system"),
                       conn->uri->path);
        return VIR_DRV_OPEN_ERROR;
    }

    if (virConnectOpenEnsureACL(conn) < 0)
        return VIR_DRV_OPEN_ERROR;

    conn->id = g_atomic_int_add(&vzct_driver->lastconnid, 1) + 1;
    conn->privateData = vzct_driver;

    return VIR_DRV_OPEN_SUCCESS;
}


static int
vzctConnectClose(virConnectPtr conn)
{
    vzctDriverPtr driver = conn->privateData;
    vzctDriverJobPtr job;

    VIR_ALLOC(job);
    job->type = VZCT_DRIVER_JOB_CONNECTION_CLOSED;
    job->conn_id = conn->id;
    if (virThreadPoolSendJob(driver->workers, 0, job) < 0) {
        VIR_WARN("cannot run cleanup job");
        vzctDriverJobFree(job);
    }

    conn->privateData = NULL;
    return 0;
}


static int
vzctDomainNetAddIpAddress(virDomainNetDefPtr net,
                          vzctl_ip_iterator ips)
{
    virSocketAddr address;
    virNetDevIPAddrPtr ip;
    unsigned int prefix = 0;
    char *p;
    char buf[512];

    if (vzctl2_env_get_ipstr(ips, buf, sizeof(buf))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("can not get ip address string for net '%s'"),
                       net->ifname_guest);
        return -1;
    }

    /* The above function returns ip in format address/mask */
    if ((p = strchr(buf, '/')))
        *p = '\0';

    if (virSocketAddrParse(&address, buf, AF_UNSPEC) < 0)
        return -1;

    if (p) {
        p++;
        if (VIR_SOCKET_ADDR_IS_FAMILY(&address, AF_INET6)) {
            /* mask is in prefix form for ipv6 */
            if (virStrToLong_uip(p, NULL, 10, &prefix) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("cannot parse prefix '%s'"), p);
                return -1;
            }
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&address, AF_INET)) {
            virSocketAddr netmask;
            int iprefix;

            /* mask is in address form for ipv4 */
            if (virSocketAddrParseIPv4(&netmask, p) < 0)
                return -1;

            if ((iprefix = virSocketAddrGetNumNetmaskBits(&netmask)) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("netmask '%s' is not valid"), p);
                return -1;
            }

            prefix = iprefix;
        }
    }

    VIR_ALLOC(ip);
    ip->address = address;
    ip->prefix = prefix;
    VIR_APPEND_ELEMENT(net->guestIP.ips, net->guestIP.nips, ip);

    return 0;
}


static void
vzctLoadNetsConfig(vzctl_env_handle_ptr env,
                   virDomainDefPtr def)
{
    vzctl_env_param_ptr param = vzctl2_get_env_param(env);
    vzctl_veth_dev_iterator it = NULL;
    vzctl_ip_iterator ips;
    virDomainNetDefPtr net;
    size_t nnets_max = 0;

    VIR_ALLOC(net);
    net->type = VIR_DOMAIN_NET_TYPE_VENET0;
    net->ifname_guest = g_strdup(VZCT_NET_VENET0);
    VIR_RESIZE_N(def->nets, nnets_max, def->nnets, 1);
    def->nets[def->nnets++] = net;

    ips = NULL;
    while ((ips = vzctl2_env_get_ipaddress(param, ips)))
        vzctDomainNetAddIpAddress(net, ips);

    while ((it = vzctl2_env_get_veth(param, it))) {
        struct vzctl_veth_dev_param net_param;

        vzctl2_env_get_veth_param(it, &net_param, sizeof(net_param));

        VIR_ALLOC(net);
        net->ifname = g_strdup(net_param.dev_name);
        net->ifname_guest = g_strdup(net_param.dev_name_ve);
        /* TODO do we need to handle invalid mac? */
        ignore_value(virMacAddrParse(net_param.mac, &net->mac));
        switch (net_param.nettype) {
        case VZCTL_NETTYPE_BRIDGE:
            net->type = VIR_DOMAIN_NET_TYPE_BRIDGE;
            net->data.bridge.brname = g_strdup(net_param.network);
            break;

        case VZCTL_NETTYPE_VNET:
            /* TODO We need to expand network to bridge for running
             * domains. */
            net->type = VIR_DOMAIN_NET_TYPE_NETWORK;
            net->data.network.name = g_strdup(net_param.network);
            break;

        default:
            /* ignore unknown types */
            VIR_DEBUG("Unknown nettype: %d", net_param.nettype);
            virDomainNetDefFree(net);
            continue;
        }
        if (net_param.vporttype == VZCTL_VPORTTYPE_OVS) {
            VIR_ALLOC(net->virtPortProfile);
            net->virtPortProfile->virtPortType = VIR_NETDEV_VPORT_PROFILE_OPENVSWITCH;
        }

        ips = NULL;
        while ((ips = vzctl2_env_get_veth_ipaddress(it, ips)))
            vzctDomainNetAddIpAddress(net, ips);

        VIR_RESIZE_N(def->nets, nnets_max, def->nnets, 1);
        def->nets[def->nnets++] = net;
    }
    VIR_SHRINK_N(def->nets, nnets_max, nnets_max - def->nnets);
}


static void
vzctLoadDisksConfig(vzctDriverPtr driver,
                    vzctl_env_handle_ptr env,
                    virDomainDefPtr def)
{
    vzctl_env_param_ptr param = vzctl2_get_env_param(env);
    struct vzctl_disk *it = NULL;
    size_t ndisks_max = 0;

    while ((it = vzctl2_env_get_disk(param, it))) {
        struct vzctl_disk_param disk_param;
        virDomainDiskDefPtr disk = virDomainDiskDefNew(driver->xmlopt);
        vzctDomainDiskPrivatePtr diskPrivate =
            (vzctDomainDiskPrivatePtr)disk->privateData;

        vzctl2_env_get_disk_param(it, &disk_param, sizeof(disk_param));

        disk->device = VIR_DOMAIN_DISK_DEVICE_DISK;
        virDomainDiskSetType(disk, VIR_STORAGE_TYPE_FILE);
        virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_PLOOP);
        virDomainDiskSetDriver(disk, "vzct");
        virDomainDiskSetSource(disk, disk_param.path);
        memcpy(diskPrivate->uuid, disk_param.uuid, sizeof(diskPrivate->uuid));

        if (STREQ_NULLABLE(disk_param.mnt, "/"))
            disk->info.bootIndex = 1;

        VIR_RESIZE_N(def->disks, ndisks_max, def->ndisks, 1);
        def->disks[def->ndisks++] = disk;
    }
    VIR_SHRINK_N(def->disks, ndisks_max, ndisks_max - def->ndisks);
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(vzctl_env_handle, vzctl2_env_close);

static void
vzctLoadDisksIotunes(vzctl_env_param_ptr param,
                     virDomainDefPtr def)
{
    unsigned int io, iops;
    size_t i;

    vzctl2_env_get_iolimit(param, &io);
    vzctl2_env_get_iopslimit(param, &iops);

    if (io == UINT_MAX)
        io = 0;
    if (iops == UINT_MAX)
        iops = 0;

    if (!io && !iops)
        return;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        virDomainBlockIoTuneInfoPtr iotune = &disk->blkdeviotune;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        iotune->total_bytes_sec = io;
        iotune->total_iops_sec = iops;

        if (def->ndisks > 1)
            iotune->group_name = g_strdup("virtuozzo");
    }
}


static virDomainDefPtr
vzctLoadDomainConfig(vzctDriverPtr driver G_GNUC_UNUSED,
                     const char *ctid,
                     char **bundlePath)
{
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainDefPtr ret = NULL;
    vzctl_env_param_ptr param;
    struct vzctl_mem_guarantee mem_guarantee;
    const char *uuid = NULL;
    const char *name = NULL;
    const char *path;
    char mask[512];
    g_autoptr(virBitmap) nodeset = NULL;
    unsigned long v;
    int err, rc;

    VZCT_CHECK_PTR_RET(NULL, env, vzctl2_env_open(ctid, 0, &err));
    param = vzctl2_get_env_param(env);
    def = virDomainDefNew();

    vzctl2_env_get_uuid(param, &uuid);
    if (!uuid || virUUIDParse(uuid, def->uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("failed to parse uuid '%s' from vzctl2_env_get_uuid"),
                       NULLSTR(uuid));
        return NULL;
    }
    vzctl2_env_get_name(env, &name);
    if (name)
        def->name = g_strdup(name);
    else
        def->name = g_strdup(ctid);
    if (bundlePath) {
        VZCT_CHECK_RET(NULL, vzctl2_env_get_ve_private_path(param, &path));
        *bundlePath = g_strdup(path);
    }
    def->extraId = g_strdup(ctid);

    def->virtType = VIR_DOMAIN_VIRT_VZCT;
    def->os.type = VIR_DOMAIN_OSTYPE_EXE;
    def->os.init = g_strdup("/sbin/init");

    vzctl2_env_get_ramsize(param, &v);
    def->mem.cur_balloon = ((unsigned long long)v) << 10;
    virDomainDefSetMemoryTotal(def, def->mem.cur_balloon);

    VZCT_CHECK_RET(NULL, vzctl2_env_get_memguarantee(param, &mem_guarantee));
    switch (mem_guarantee.type) {
    case VZCTL_MEM_GUARANTEE_PCT:
        if (mem_guarantee.value > 100)
            mem_guarantee.value = 100;
        def->mem.min_guarantee = def->mem.cur_balloon * mem_guarantee.value / 100;
        break;

    case VZCTL_MEM_GUARANTEE_BYTES:
        /* bytes to KiB */
        def->mem.min_guarantee = mem_guarantee.value >> 10;
        break;

    case VZCTL_MEM_GUARANTEE_AUTO:
        def->mem.min_guarantee_auto = true;
        break;

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("unexpected memguarantee type=%d"),
                       mem_guarantee.type);
        return NULL;
    }

    vzctl2_env_get_cpu_count(param, &v);
    def->maxvcpus = v;

    rc = vzctl2_env_get_cpumask(param, mask, sizeof(mask));
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_cpumask);
        return NULL;
    }
    if (rc == 0 && strlen(mask) > 0 &&
        virBitmapParse(mask, &def->cpumask, VIR_DOMAIN_CPUMASK_LEN) < 0)
        return NULL;

    rc = vzctl2_env_get_nodemask(param, mask, sizeof(mask));
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_nodemask);
        return NULL;
    }
    if (rc == 0 && strlen(mask) > 0) {
        if (virBitmapParse(mask, &nodeset, VIR_DOMAIN_CPUMASK_LEN) < 0)
            return NULL;

        if (virDomainNumatuneSet(def->numa,
                                 true,
                                 VIR_DOMAIN_NUMATUNE_PLACEMENT_STATIC,
                                 VIR_DOMAIN_NUMATUNE_MEM_STRICT,
                                 nodeset) < 0)
            return NULL;
    }

    def->onReboot = VIR_DOMAIN_LIFECYCLE_ACTION_RESTART;
    def->onPoweroff = VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY;
    def->onCrash = VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY;

    vzctLoadDisksConfig(driver, env, def);
    vzctLoadDisksIotunes(param, def);
    vzctLoadNetsConfig(env, def);

    ret = g_steal_pointer(&def);
    return ret;
}


static bool
vzctGenerateDiskNames(virDomainDefPtr def)
{
    size_t i;
    int max = -1;
    bool save = false;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        int idx;

        /*
         * All disks added NOT thru libvirt will be set on virtio
         * bus. This is handy for SolusIO. In future we can make
         * default bus configurable if we need to.
         */
        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK ||
            !disk->dst ||
            !STRPREFIX(disk->dst, "vd"))
            continue;

        idx = virDiskNameToIndex(disk->dst);
        if (idx > max)
            max = idx;
    }

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK ||
            disk->dst)
            continue;

        disk->dst = virIndexToDiskName(++max, "vd");
        disk->bus = VIR_DOMAIN_DISK_BUS_VIRTIO;
        save = true;
    }

    return save;
}


static bool
vzctMergeConfigs(virDomainDefPtr def, virDomainDefPtr drvDef)
{
    size_t i;

    if (!drvDef)
        return vzctGenerateDiskNames(def);

    def->id = drvDef->id;

    if (drvDef->ngraphics) {
        def->ngraphics = drvDef->ngraphics;
        drvDef->ngraphics = 0;
        def->graphics = g_steal_pointer(&drvDef->graphics);
    }

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        vzctDomainDiskPrivatePtr diskPrivate =
            (vzctDomainDiskPrivatePtr)disk->privateData;
        size_t j;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        for (j = 0; j < drvDef->ndisks; j++) {
            virDomainDiskDefPtr drvDisk = drvDef->disks[j];
            vzctDomainDiskPrivatePtr drvDiskPrivate =
                (vzctDomainDiskPrivatePtr)drvDisk->privateData;

            if ((*drvDiskPrivate->uuid &&
                 STREQ_NULLABLE(diskPrivate->uuid, drvDiskPrivate->uuid)) ||
                (!*drvDiskPrivate->uuid &&
                 STREQ_NULLABLE(disk->src->path, drvDisk->src->path))) {
                disk->dst = g_strdup(drvDisk->dst);
                disk->bus = drvDisk->bus;
                disk->info.alias = g_strdup(drvDisk->info.alias);

                disk->src->type = drvDisk->src->type;
                disk->src->srcpool = g_steal_pointer(&drvDisk->src->srcpool);
                break;
            }
        }
    }

    for (i = 0; i < drvDef->ndisks; i++) {
        if (drvDef->disks[i]->device == VIR_DOMAIN_DISK_DEVICE_CDROM) {
            VIR_EXPAND_N(def->disks, def->ndisks, 1);
            /*
             * This will make a hole in drvDef but the latter is not supposed
             * to be used after call to this function anyway.
             */
            def->disks[def->ndisks - 1] = g_steal_pointer(&drvDef->disks[i]);
            break;
        }

    }

    for (i = 0; i < def->nnets; i++) {
        virDomainNetDef *net = def->nets[i];
        size_t j;

        for (j = 0; j < drvDef->nnets; j++) {
            const virDomainNetDef *drvNet = drvDef->nets[j];

            /* venet0 does not have a device on host */
            if ((net->ifname && drvNet->ifname &&
                 STREQ(net->ifname, drvNet->ifname)) ||
                (net->ifname_guest && drvNet->ifname_guest &&
                 STREQ(net->ifname_guest, drvNet->ifname_guest))) {
                net->info.alias = g_strdup(drvNet->info.alias);
                break;
            }
        }
    }

    def->metadata = g_steal_pointer(&drvDef->metadata);

    return vzctGenerateDiskNames(def);
}


#define VZCT_VNC_SERVER_BIN "/usr/bin/prl_vzvncserver_app"

static int
vzctStartVNCServer(vzctDriverPtr driver,
                   virDomainObjPtr dom)
{
    g_autofree char *pidfile = vzctVncPidFile(driver->config, dom->def->uuid);
    vzctDomainObjPrivatePtr priv = dom->privateData;
    virDomainGraphicsDefPtr gr;
    g_autoptr(virCommand) cmd = NULL;

    if (!dom->def->ngraphics)
        return 0;

    gr = dom->def->graphics[0];
    if (!virFileIsExecutable(VZCT_VNC_SERVER_BIN)) {
        virReportSystemError(errno,
                             _("Cannot find %s - "
                               "Possibly the package isn't installed"),
                             VZCT_VNC_SERVER_BIN);
        return -1;
    }

    if (gr->data.vnc.autoport) {
        unsigned short port;

        if (virPortAllocatorAcquire(driver->vncPorts, &port) < 0)
            return -1;

        gr->data.vnc.port = port;
    } else if (virPortAllocatorSetUsed(gr->data.vnc.port) < 0) {
        return -1;
    }
    gr->data.vnc.portReserved = true;

    cmd = virCommandNew(VZCT_VNC_SERVER_BIN);

    virCommandAddArgList(cmd, "--listen", gr->listens[0].address, NULL);
    virCommandAddArgFormat(cmd, "--port=%d", gr->data.vnc.port);

    if (gr->data.vnc.auth.passwd) {
        virCommandAddArg(cmd, "--passwd");
        virCommandSetInputBuffer(cmd, gr->data.vnc.auth.passwd);
    }

    virCommandAddArg(cmd, dom->def->name);

    if (unlink(pidfile) < 0 && errno != ENOENT) {
        virReportSystemError(errno,
                             _("Cannot remove stale VNC PID file %s"), pidfile);
        return -1;
    }

    virCommandSetPidFile(cmd, pidfile);
    virCommandDaemonize(cmd);

    if (virCommandRun(cmd, NULL) < 0)
        return -1;

    if (virPidFileReadPath(pidfile, &priv->vncPid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Failed to read vnc server pid"));
        return -1;
    }

    return 0;
}


static void
vzctStopVNCServer(vzctDriverPtr driver, virDomainObjPtr dom)
{
    g_autofree char *pidfile = vzctVncPidFile(driver->config, dom->def->uuid);
    vzctDomainObjPrivatePtr priv = dom->privateData;
    virDomainGraphicsDefPtr gr;

    if (!dom->def->ngraphics)
        return;

    gr = dom->def->graphics[0];
    /* FIXME Check pid belongs to process of VZCT_VNC_SERVER_BIN. */
    if (priv->vncPid) {
        virProcessKillPainfully(priv->vncPid, true);
        priv->vncPid = 0;
    }
    if (gr->data.vnc.portReserved)
        virPortAllocatorRelease(gr->data.vnc.port);

    if (unlink(pidfile) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove VNC PID file: %s, errno: %d", pidfile, errno);
}


static void
vzctDomainFinishExecProcs(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    size_t i;

    for (i = 0; i < priv->exec_nprocs; i++) {
        vzctDomainExecProcessPtr exec_proc = priv->exec_procs[i];
        int status;
        int rc;

        if (!exec_proc->terminated)
            vzctl2_env_exec_terminate(exec_proc->handle);

        VIR_DEBUG("Reaping process pid=%d", exec_proc->handle->pid);
        rc = vzctl2_env_waitpid(exec_proc->handle, 0, &status);
        if (rc < 0)
            VIR_WARN("Fail reaping process pid=%d, err: %s",
                     exec_proc->handle->pid, NULLSTR(vzctl2_get_last_error()));

        vzctl2_release_exec_handle(exec_proc->handle);
    }

    virEventRemoveTimeout(priv->cleanupTimer);
    priv->cleanupTimer = 0;
    VIR_FREE(priv->exec_procs);
    priv->exec_nprocs = 0;
    priv->exec_nprocs_max = 0;
}


static void
vzctDomainStopFlags(vzctDriverPtr driver,
                    virDomainObjPtr dom,
                    bool skipEnv)
{
    g_autofree char *statefile = vzctStateFile(driver->config, dom->def->uuid);
    g_autoptr(vzctl_env_handle) env = NULL;
    int err;

    if (!skipEnv && !(env = vzctl2_env_open(dom->def->extraId, 0, &err)))
        VZCT_ERROR_MESSAGE(vzctl2_env_open);

    vzctDomainFinishExecProcs(dom);
    vzctStopVNCServer(driver, dom);
    if (env)
        vzctDomainClearVolumes(dom->def, env);
    virDomainObjRemoveTransientDef(dom);
    virDomainObjSetState(dom, VIR_DOMAIN_SHUTOFF, -1);

    if (unlink(statefile) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove state, file: %s, error: %d",
                 statefile, errno);
}


static void
vzctDomainStop(vzctDriverPtr driver,
               virDomainObjPtr dom)
{
    vzctDomainStopFlags(driver, dom, false);
}


static int
vzctDomainStart(vzctDriverPtr driver,
                virDomainObjPtr dom,
                virDomainDefPtr drvDef)
{
    g_autofree char *statefile = vzctStateFile(driver->config, dom->def->uuid);
    g_autoptr(virDomainDef) drvDefCopy = NULL;
    vzctDomainObjPrivatePtr priv = dom->privateData;
    virDomainDefPtr def;
    int ret = -1;

    /* FIXME Use active config from vzctl when it will be available. */
    if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
        return -1;
    dom->newDef = g_steal_pointer(&dom->def);
    dom->def = g_steal_pointer(&def);

    if (!drvDef)
        drvDef = dom->newDef;
    /* Make a copy as merge damages second argument. */
    if (!(drvDefCopy = vzctDomainDefCopy(drvDef, driver->xmlopt, false)))
        goto cleanup;
    vzctMergeConfigs(dom->def, drvDefCopy);

    virDomainObjSetState(dom, VIR_DOMAIN_RUNNING, -1);
    dom->def->id = 0;

    if (vzctStartVNCServer(driver, dom) < 0)
        goto cleanup;

    virObjectRef(dom);
    if ((priv->cleanupTimer = virEventAddTimeout(-1,
                                                 vzctDomainCleanupTimeout,
                                                 dom,
                                                 virObjectFreeCallback)) < 0) {
        virObjectUnref(dom);
        goto cleanup;
    }

    if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    if (ret < 0)
        vzctDomainStop(driver, dom);

    return ret;
}


static virDomainObjPtr
vzctAddDomain(vzctDriverPtr driver,
              const char *ctid,
              virDomainObjPtr dom)
{
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(virDomainDef) drvDef = NULL;
    vzctl_env_status_t status;
    virDomainState state = VIR_DOMAIN_SHUTOFF;
    vzctDomainObjPrivatePtr priv;
    g_autofree char *path = NULL;
    g_autofree char *statefile = NULL;
    g_autofree char *migfile = NULL;
    g_autoptr(virDomainObj) mig = NULL;
    g_autofree char *autostartFile = NULL;
    bool failure = true;
    int rc = 0;

    if (!dom) {
        g_autofree char *bundlePath = NULL;

        if (!(def = vzctLoadDomainConfig(driver, ctid, &bundlePath)))
            goto cleanup;

        if (!(dom = virDomainObjListAdd(driver->domains, def, driver->xmlopt, 0)))
            goto cleanup;

        /* When we create new domain thru defineXML ADDED event will be fired
         * before creating is finished thus the def readed will be not complete.
         * So after creation is finished and dom unlocked we update config
         * to the outdated version. Thus let's skip update in this case.
         * May be we just should create and config domain in one go in defineXML.
         */
        if (dom->def)
            return dom;

        priv = dom->privateData;
        priv->bundlePath = g_steal_pointer(&bundlePath);
        dom->persistent = true;

        vzctDomainLoadSnapshotMetadata(driver, dom);
    } else {
        priv = dom->privateData;
        VIR_FREE(dom->newDef);

        if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
            goto cleanup;
    }
    VIR_FREE(dom->def);
    dom->def = g_steal_pointer(&def);
    dom->def->id = -1;
    virDomainObjSetState(dom, VIR_DOMAIN_SHUTOFF, -1);

    path = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/config.xml", priv->bundlePath);
    if (virFileExists(path)) {
        unsigned int parseFlags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                                  VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                                  VIR_DOMAIN_DEF_PARSE_ALLOW_POST_PARSE_FAIL |
                                  VIR_DOMAIN_DEF_PARSE_DISK_PRIVATE;
        if (!(drvDef = virDomainDefParseFile(path, driver->xmlopt,
                                             NULL, parseFlags)))
            goto cleanup;
    }
    if (vzctMergeConfigs(dom->def, drvDef) &&
        vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_RUNNING));
    /* TODO examine state more carefully */
    VIR_DEBUG("container: %s, status: %d", dom->def->name, status.mask);
    if (status.mask & ENV_STATUS_RUNNING)
        state = VIR_DOMAIN_RUNNING;

    migfile = vzctMigFile(dom);
    if (virFileExists(migfile)) {
        if (!(mig = virDomainObjParseFile(migfile, driver->xmlopt,
                                          VIR_DOMAIN_DEF_PARSE_STATUS |
                                          VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE)))
            goto cleanup;
        virObjectUnlock(mig);
    }

    statefile = vzctStateFile(driver->config, dom->def->uuid);
    if (virFileExists(statefile))
        rc = vzctDomainStateLoad(driver, dom, dom->def->uuid);
    else if (state == VIR_DOMAIN_RUNNING)
        rc = vzctDomainStart(driver, dom, mig ? mig->def : NULL);
    if (rc < 0)
        goto cleanup;
    if (state == VIR_DOMAIN_SHUTOFF &&
        dom->state.state == VIR_DOMAIN_RUNNING) {
        vzctDomainStop(driver, dom);
    }

    if (mig && unlink(migfile) < 0)
        VIR_WARN("Cannot remove migrating status file: %s, error: %d",
                 migfile, errno);

    autostartFile = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/autostart",
                                    priv->bundlePath);
    if (virFileExists(autostartFile))
        dom->autostart = true;

    failure = false;

 cleanup:
    if (dom && failure) {
        VIR_WARN("Can not sync state for domain ctid=%s", dom->def->extraId);
        virDomainObjSetState(dom, VIR_DOMAIN_NOSTATE, -1);
        dom->def->id = -1;
    }

    return dom;
}


/*
 * Try to sync domain if it is out of sync.
 * Returns:
 *  1 - domain is in sync
 * -1 - sync error
 *  0 - domain synced sucessfully
 */
static int
vzctCheckSync(vzctDriverPtr driver, virDomainObjPtr dom)
{
    g_autofree char *oldname = NULL;
    virObjectEventPtr event;

    if (dom->state.state != VIR_DOMAIN_NOSTATE)
        return 1;

    oldname = g_strdup(dom->def->name);

    vzctAddDomain(driver, NULL, dom);
    if (dom->state.state == VIR_DOMAIN_NOSTATE) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("domain is out of sync"));
        return -1;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    if (STRNEQ(dom->def->name, oldname) &&
        virDomainObjListRename(driver->domains, dom, dom->def->name,
                               0, NULL, NULL) < 0) {
        virDomainObjSetState(dom, VIR_DOMAIN_NOSTATE, -1);
        dom->def->id = -1;
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("domain is out of sync"));
        return -1;
    }

    return 0;
}


static int
vzctUpdateDomain(vzctDriverPtr driver,
                 virDomainObjPtr dom)
{
    g_autofree char *statefile = NULL;
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(virDomainDef) oldDef = NULL;
    g_autoptr(virDomainDef) oldNewDef = NULL;
    bool saveConfig, saveState;
    bool save;

    /*
     * If domain is active then use active config from
     * vzctl when it will be available.
     */
    if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
        return -1;

    if (STRNEQ(dom->def->name, def->name) &&
        virDomainObjListRename(driver->domains, dom, def->name,
                               0, NULL, NULL) < 0)
        return -1;
    oldDef = g_steal_pointer(&dom->def);
    save = vzctMergeConfigs(def, oldDef);
    dom->def = g_steal_pointer(&def);

    if (virDomainObjIsActive(dom)) {
        if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
            return -1;
        oldNewDef = g_steal_pointer(&dom->newDef);
        saveConfig = vzctMergeConfigs(def, oldNewDef);
        dom->newDef = g_steal_pointer(&def);
        saveState = save;
    } else {
        saveConfig = save;
        saveState = false;
    }

    if (saveConfig && vzctDomainSaveConfig(driver, dom) < 0)
        return -1;
    statefile = vzctStateFile(driver->config, dom->def->uuid);
    if (saveState &&
        virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
        return -1;

    return 0;
}


static int
vzctStateCleanup(void)
{
    vzctDriverPtr driver = vzct_driver;

    /*
     * In unpriviledged mode vzct driver initialization is skipped. Also
     * cleanup is called even on initialization failure.
     */
    if (!vzct_driver)
        return 0;

    virThreadPoolFree(driver->workers);
    vzctDriverConnectClose(driver);
    virObjectUnref(driver->xmlopt);
    virObjectUnref(driver->caps);
    virMutexDestroy(&driver->lock);
    virObjectUnref(driver->domains);
    VIR_FREE(driver->config->stateDir);
    VIR_FREE(driver->config);
    g_object_unref(driver->eventThread);
    virPortAllocatorRangeFree(driver->vncPorts);
    virObjectUnref(driver->eventsManager);
    VIR_FREE(driver);
    vzct_driver = NULL;
    return 0;
}


static void*
vzctDomainObjPrivateAlloc(void *opaque G_GNUC_UNUSED)
{
    vzctDomainObjPrivatePtr priv;

    VIR_ALLOC(priv);
    if (virCondInit(&priv->jobCond) < 0) {
        virReportSystemError(errno, "%s", _("cannot initialize job condition"));
        VIR_FREE(priv);
        return NULL;
    }
    priv->snapshots = virDomainSnapshotObjListNew();
    return priv;
}


static void
vzctDomainObjPrivateFree(void *data)
{
    vzctDomainObjPrivatePtr priv = data;

    virCondDestroy(&priv->jobCond);
    virDomainSnapshotObjListFree(priv->snapshots);
    VIR_FREE(priv->bundlePath);
    VIR_FREE(priv);
}


static int
vzctDomainDeviceDefPostParse(virDomainDeviceDefPtr dev,
                             const virDomainDef *def G_GNUC_UNUSED,
                             unsigned int parseFlags G_GNUC_UNUSED,
                             void *opaque G_GNUC_UNUSED,
                             void *parseOpaque G_GNUC_UNUSED)
{
    virDomainDiskDefPtr disk;

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        disk = dev->data.disk;

        if (!virDomainDiskGetDriver(disk))
            virDomainDiskSetDriver(disk, "vzct");

        if (!disk->src->format) {
            if (disk->device == VIR_DOMAIN_DISK_DEVICE_DISK)
                virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_PLOOP);
            else
                virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_RAW);
        }

        break;
    }

    return 0;
}


static int
vzctValidateDiskIoTune(const virDomainDef *def,
                       const virDomainDiskDef *disk)
{
    const virDomainBlockIoTuneInfo *iotune = &disk->blkdeviotune;
    size_t i;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
        if (virDomainBlockIoTuneInfoHasAny(iotune)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("iotune is only supported for disks devices"));
            return -1;
        }

        return 0;
    }

    if (iotune->read_bytes_sec ||
        iotune->write_bytes_sec ||
        iotune->read_iops_sec ||
        iotune->write_iops_sec ||
        virDomainBlockIoTuneInfoHasMax(iotune) ||
        virDomainBlockIoTuneInfoHasMaxLength(iotune)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only total_bytes_sec and total_iops_sec"
                         " can be tuned for disks"));
        return -1;
    }

    if (!iotune->group_name)
        return 0;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *d = def->disks[i];

        if (STRNEQ_NULLABLE(iotune->group_name, d->blkdeviotune.group_name))
            continue;

        if (!virDomainBlockIoTuneInfoEqual(iotune, &d->blkdeviotune)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("different iotunes for disks %s and %s"),
                           disk->dst, d->dst);
            return -1;
        }
    }

    return 0;
}


static int
vzctDeviceDefValidateCallback(const virDomainDeviceDef *dev,
                              const virDomainDef *def G_GNUC_UNUSED,
                              void *opaque G_GNUC_UNUSED)
{
    virDomainDiskDefPtr disk;
    virDomainNetDefPtr net;
    virDomainGraphicsDefPtr gr;

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        disk = dev->data.disk;
        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK &&
            disk->device != VIR_DOMAIN_DISK_DEVICE_CDROM) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only cdrom and disks are supported"));
            return -1;
        }

        if (disk->src->type != VIR_STORAGE_TYPE_FILE &&
            disk->src->type != VIR_STORAGE_TYPE_VOLUME) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only files and volume are supported for disks"));
            return -1;
        }

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_DISK &&
            disk->src->format != VIR_STORAGE_FILE_PLOOP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only ploop files are supported for disks"));
            return -1;
        }

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM &&
            disk->src->format != VIR_STORAGE_FILE_RAW) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only raw files are supported for cdroms"));
            return -1;
        }

        if (STRNEQ(virDomainDiskGetDriver(disk), "vzct")) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only vzct driver is supported for disks"));
            return -1;
        }

        if ((disk->info.bootIndex > 1 &&
             disk->device == VIR_DOMAIN_DISK_DEVICE_DISK) ||
            (disk->info.bootIndex > 0 &&
             disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only one boot disk is supported and it must"
                             " have boot index 1"));
            return -1;
        }

        if (vzctValidateDiskIoTune(def, disk) < 0)
            return -1;

        break;

    case VIR_DOMAIN_DEVICE_NET:
        net = dev->data.net;
        if (net->type != VIR_DOMAIN_NET_TYPE_BRIDGE &&
            net->type != VIR_DOMAIN_NET_TYPE_NETWORK &&
            net->type != VIR_DOMAIN_NET_TYPE_VENET0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only interfaces of type bridge, network"
                             " and venet0 are supported"));
            return -1;
        }

        if (net->type == VIR_DOMAIN_NET_TYPE_VENET0) {
            virMacAddr zeromac;

            if (net->ifname_guest &&
                STRNEQ(net->ifname_guest, VZCT_NET_VENET0)) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Guest dev name can not be other than venet0"));
                return -1;
            }

            if (net->ifname) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Host dev name can not be specified for venet0"));
                return -1;
            }

            memset(zeromac.addr, 0, sizeof(zeromac.addr));
            if (!net->mac_generated &&
                memcmp(net->mac.addr, zeromac.addr, sizeof(zeromac.addr))) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("mac can not be specified for venet0"));
                return -1;
            }

            break;
        }

        if (!net->ifname_guest) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Guest dev name must be specified"));
            return -1;
        }
        if (net->virtPortProfile &&
            net->virtPortProfile->virtPortType != VIR_NETDEV_VPORT_PROFILE_OPENVSWITCH) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only explicit openvswitch virtual port type"
                             " is supported"));
            return -1;
        }
        break;

    case VIR_DOMAIN_DEVICE_GRAPHICS:
        gr = dev->data.graphics;
        if (gr->type != VIR_DOMAIN_GRAPHICS_TYPE_VNC) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only VNC graphics is supported."));
            return -1;
        }
        if (gr->data.vnc.websocket != 0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Websocket VNC ports are not supported."));
            return -1;
        }
        if (gr->data.vnc.keymap != 0 &&
            STRNEQ(gr->data.vnc.keymap, "en-us")) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only 'en-us' VNC keymap is supported."));
            return -1;
        }
        if (gr->data.vnc.sharePolicy == VIR_DOMAIN_GRAPHICS_VNC_SHARE_ALLOW_EXCLUSIVE) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("VNC exclusive share policy is not supported."));
            return -1;
        }
        if (gr->data.vnc.auth.connected == VIR_DOMAIN_GRAPHICS_AUTH_CONNECTED_FAIL ||
            gr->data.vnc.auth.connected == VIR_DOMAIN_GRAPHICS_AUTH_CONNECTED_KEEP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("VNC supports only disconnecton on passwd change."));
            return -1;
        }
        if (gr->data.vnc.auth.expires) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("VNC expiring passwords is not supported."));
            return -1;
        }
        if (gr->nListens > 1) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only one graphics listen address is supported."));
            return -1;
        }
        if (gr->listens[0].type != VIR_DOMAIN_GRAPHICS_LISTEN_TYPE_ADDRESS) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only listening on inet address is supported for graphics"));
            return -1;
        }
    }

    return 0;
}


static int
vzctDomainDefValidateIoTune(const virDomainDef *def)
{
    size_t i;
    int tuneCount = 0;
    int groupCount = 0;
    int diskCount = 0;
    const char *group_name = NULL;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *disk = def->disks[i];
        const virDomainBlockIoTuneInfo *iotune  = &disk->blkdeviotune;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        diskCount++;

        if (!virDomainBlockIoTuneInfoHasAny(iotune))
            continue;

        if (iotune->group_name) {
            if (group_name) {
                if (STRNEQ(group_name, iotune->group_name))
                    goto error;
            } else {
                group_name = iotune->group_name;
            }
            groupCount++;
        }

        tuneCount++;
    }

    if (!tuneCount ||
        (tuneCount == 1 && diskCount == 1) ||
        (group_name && groupCount == diskCount))
        return 0;

 error:
    virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                   _("Supported configurations are either single disk"
                     " with iotunes or all disks in same iotune group"));
    return -1;
}


static int
vzctDomainDefValidateCallback(const virDomainDef *def,
                              void *opaque G_GNUC_UNUSED)
{
    size_t i;
    int cdroms = 0;
    int venet0 = 0;
    virDomainNumatuneMemMode mode;

    if (def->onReboot != VIR_DOMAIN_LIFECYCLE_ACTION_RESTART) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only restart on reboot is supported"));
        return -1;
    }
    if (def->onPoweroff != VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only destroy on poweroff is supported"));
        return -1;
    }
    if (def->onCrash != VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only destroy on crash is supported"));
        return -1;
    }
    if (virDomainDefHasVcpusOffline(def)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Offline vcpus are not supported"));
        return -1;
    }
    if (def->placement_mode != VIR_DOMAIN_CPU_PLACEMENT_MODE_STATIC) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only static cpu placement mode is supported"));
        return -1;
    }
    if (virDomainNumatuneGetMode(def->numa, -1, &mode) == 0 &&
        mode != VIR_DOMAIN_NUMATUNE_MEM_STRICT) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only strict numatune mode is supported"));
        return -1;
    }
    if (virDomainNumatuneHasPlacementAuto(def->numa)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Auto numatune placement mode is not supported"));
        return -1;
    }
    if (def->mem.cur_balloon != virDomainDefGetMemoryTotal(def)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Memory balloning is not supported"));
        return -1;
    }
    if (STRNEQ("/sbin/init", def->os.init)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("init must be /sbin/init"));
        return -1;
    }
    if (def->ngraphics > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only one graphics per domain is supported."));
        return -1;
    }

    for (i = 0; i < def->ndisks; i++) {
        if (def->disks[i]->device == VIR_DOMAIN_DISK_DEVICE_CDROM)
            cdroms++;
    }

    if (cdroms > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only one cdrom is supported."));
        return -1;
    }

    for (i = 0; i < def->nnets; i++) {
        if (def->nets[i]->type == VIR_DOMAIN_NET_TYPE_VENET0)
            venet0++;
    }

    if (venet0 > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Domain can only have single venet0."));
        return -1;
    }

    if (vzctDomainDefValidateIoTune(def) < 0)
        return -1;

    return 0;
}


static int
vzctDomainDiskParse(xmlXPathContextPtr ctxt,
                    virDomainDiskDefPtr disk)
{
    vzctDomainDiskPrivatePtr diskPrivate =
        (vzctDomainDiskPrivatePtr)disk->privateData;
    g_autofree char *uuid = NULL;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
        return 0;

    if (!(uuid = virXPathString("string(./uuid[1])", ctxt)) ||
        virStrcpyStatic(diskPrivate->uuid, uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("cannot parse uuid"));
        return -1;
    }

    return 0;
}


static int
vzctDomainDiskFormat(virDomainDiskDefPtr disk,
                     virBufferPtr buf)
{
    vzctDomainDiskPrivatePtr diskPrivate =
        (vzctDomainDiskPrivatePtr)disk->privateData;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
        return 0;

    virBufferAsprintf(buf, "<uuid>%s</uuid>\n", diskPrivate->uuid);
    return 0;
}

virDomainDefParserConfig vzctDriverDomainDefParserConfig = {
    .devicesPostParseCallback = vzctDomainDeviceDefPostParse,
    .deviceValidateCallback = vzctDeviceDefValidateCallback,
    .domainValidateCallback = vzctDomainDefValidateCallback,
    .features = VIR_DOMAIN_DEF_FEATURE_USER_ALIAS,
};

virDomainXMLPrivateDataCallbacks vzctDriverPrivateDataCallbacks = {
    .alloc = vzctDomainObjPrivateAlloc,
    .free = vzctDomainObjPrivateFree,
    .diskNew = vzctDomainDiskPrivateNew,
    .diskParse = vzctDomainDiskParse,
    .diskFormat = vzctDomainDiskFormat,
};


static virCapsPtr
vzctDriverCapsInit(void)
{
    virCapsPtr caps;
    virCapsGuestPtr guest;

    if ((caps = virCapabilitiesNew(virArchFromHost(),
                                   false, false)) == NULL)
        goto error;

    if (!(caps->host.numa = virCapabilitiesHostNUMANewHost()))
        goto error;

    if (virGetHostUUID(caps->host.host_uuid)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       "%s", _("cannot get the host uuid"));
        goto error;
    }

    if ((guest = virCapabilitiesAddGuest(caps,
                                         VIR_DOMAIN_OSTYPE_EXE,
                                         caps->host.arch,
                                         NULL, NULL, 0, NULL)) == NULL)
        goto error;

    if (virCapabilitiesAddGuestDomain(guest,
                                      VIR_DOMAIN_VIRT_VZCT,
                                      NULL, NULL, 0, NULL) == NULL)
        goto error;

    return caps;

 error:
    virObjectUnref(caps);
    return NULL;
}


static void
vzctDriverJobHandleEvent(vzctDriverPtr driver,
                         int event,
                         const char *ctid)
{
    virDomainObjPtr dom;
    virObjectEventPtr domEvent;
    bool job = false;
    int rc;

    if (event == VZCT_DOMAIN_EVENT_DEFINED) {
        dom = vzctAddDomain(driver, ctid, NULL);
        if (dom) {
            if (virDomainObjIsActive(dom)) {
                domEvent = virDomainEventLifecycleNewFromObj(dom,
                                             VIR_DOMAIN_EVENT_STARTED,
                                             VIR_DOMAIN_EVENT_STARTED_BOOTED);
                virObjectEventStateQueue(driver->eventsManager, domEvent);
            }
            domEvent = virDomainEventLifecycleNewFromObj(dom,
                                             VIR_DOMAIN_EVENT_DEFINED,
                                             VIR_DOMAIN_EVENT_DEFINED_ADDED);
            virObjectEventStateQueue(driver->eventsManager, domEvent);
        }
        goto cleanup;
    }

    if (!(dom = virDomainObjListFindByExtraId(driver->domains, ctid))) {
        /* Not an error - container can be deleted at the moment
         * this event is arrived */
        VIR_DEBUG("Can not find container by ctid: %s", ctid);
        return;
    }

    if (vzctDomainBeginJobInternal(dom, VZCT_DOMAIN_JOB_MODIFY, -1) < 0)
        goto cleanup;
    job = true;

    if (event == VZCT_DOMAIN_EVENT_UNDEFINED) {
        virDomainObjListRemove(driver->domains, dom);
        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_UNDEFINED,
                                         VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        goto cleanup;
    }

    rc = vzctCheckSync(driver, dom);
    if (rc < 0 || (rc == 0 && event == VZCT_DOMAIN_EVENT_UPDATED))
        goto cleanup;

    switch (event) {
    case VZCT_DOMAIN_EVENT_STARTED:
        if (virDomainObjIsActive(dom))
            break;

        if (vzctDomainStart(driver, dom, NULL) < 0)
            goto error;

        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_STARTED,
                                         VIR_DOMAIN_EVENT_STARTED_BOOTED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        break;

    case VZCT_DOMAIN_EVENT_STOPPED:
        if (!virDomainObjIsActive(dom))
            break;

        vzctDomainStop(driver, dom);
        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_STOPPED,
                                         VIR_DOMAIN_EVENT_STOPPED_DESTROYED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        break;

    case VZCT_DOMAIN_EVENT_UPDATED:
        if (vzctUpdateDomain(driver, dom) < 0)
            goto error;
        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_DEFINED,
                                         VIR_DOMAIN_EVENT_DEFINED_UPDATED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        break;
    }

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return;

 error:
    VIR_WARN("Can not sync state for domain ctid=%s", ctid);
    virDomainObjSetState(dom, VIR_DOMAIN_NOSTATE, -1);
    dom->def->id = -1;
    domEvent = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, domEvent);
    goto cleanup;
}


static void
vzctDomainCleanupTimeout(int timer, void *opaque)
{
    vzctDriverJobPtr job;

    VIR_ALLOC(job);
    job->type = VZCT_DRIVER_JOB_CLEANUP;
    job->dom = opaque;
    if (virThreadPoolSendJob(vzct_driver->workers, 0, job) < 0) {
        VIR_WARN("cannot run cleanup job");
        vzctDriverJobFree(job);
    }
    virEventUpdateTimeout(timer, -1);
}


static void
vzctDomainCleanup(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    size_t i;

    VIR_DEBUG("Cleanup domain name=%s", dom->def->name);

    for (i = 0; i < priv->exec_nprocs; i++) {
        vzctDomainExecProcessPtr exec_proc = priv->exec_procs[i];
        int status;
        int rc;

        if (!exec_proc->orphan)
            continue;

        if (!exec_proc->terminated) {
            vzctl2_env_exec_terminate(exec_proc->handle);
            exec_proc->terminated = true;
        }

        VIR_DEBUG("Reaping process pid=%d", exec_proc->handle->pid);
        rc = vzctl2_env_waitpid(exec_proc->handle, 1, &status);
        if (rc == 0)
            continue;

        if (rc < 0)
            VIR_WARN("Fail reaping process pid=%d, err: %s",
                     exec_proc->handle->pid, NULLSTR(vzctl2_get_last_error()));

        /* pid is reset in handle after reaping is finished */
        VIR_DEBUG("Process reaped.");
        vzctl2_release_exec_handle(exec_proc->handle);
        VIR_DELETE_ELEMENT_INPLACE(priv->exec_procs, i, priv->exec_nprocs);
    }

    for (i = 0; i < priv->exec_nprocs; i++) {
        if (priv->exec_procs[i]->orphan) {
            VIR_DEBUG("Cleanup is not finished. Retry after timeout.");
            virEventUpdateTimeout(priv->cleanupTimer, 30 * 1000);
            break;
        }
    }
}


static void
vzctDriverJobHandleConnectionClosed(vzctDriverPtr driver,
                                    int conn_id)
{
    virDomainObjPtr *doms;
    size_t ndoms;
    size_t i;
    size_t j;

    /*
     * Like for qemu driver collect domains instead of using iterator.
     * Because if we stuck in iterator for some domain then read lock
     * for the list is hold and define/undefine API for example for
     * another domain will stuck which is not good.
     */
    virDomainObjListCollect(driver->domains, NULL, &doms, &ndoms, NULL, 0);
    VIR_DEBUG("run connection closed hanlder");

    for (i = 0; i < ndoms; i++) {
        virDomainObjPtr dom = doms[i];
        vzctDomainObjPrivatePtr priv = dom->privateData;

        virObjectLock(dom);

        if (priv->job == VZCT_DOMAIN_JOB_MIGRATION &&
            priv->mig_conn_id == conn_id)
            vzctDomainDropJob(dom);

        for (j = 0; j < priv->exec_nprocs; j++) {
            vzctDomainExecProcessPtr exec_proc = priv->exec_procs[j];

            if (exec_proc->conn_id == conn_id)
                exec_proc->orphan = true;
        }

        vzctDomainCleanup(dom);
        virObjectUnlock(dom);
    }
}


static void
vzctDriverJobHandler(void *data, void *opaque)
{
    vzctDriverJobPtr job = data;
    vzctDriverPtr driver = opaque;

    switch (job->type) {
    case VZCT_DRIVER_JOB_EVENT:
        vzctDriverJobHandleEvent(driver, job->eventType, job->ctid);
        break;
    case VZCT_DRIVER_JOB_CONNECTION_CLOSED:
        vzctDriverJobHandleConnectionClosed(driver, job->conn_id);
        break;
    case VZCT_DRIVER_JOB_CLEANUP:
        virObjectLock(job->dom);
        vzctDomainCleanup(job->dom);
        virObjectUnlock(job->dom);
        break;
    }

    vzctDriverJobFree(job);
}


static int
vzctDomainLoadSnapshotMetadata(vzctDriverPtr driver,
                               virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    g_autofree char *dirpath = NULL;
    struct dirent *entry;
    DIR *dir = NULL;

    dirpath = g_strdup_printf("%s/"VZCT_SNAPHOTS_DIR, priv->bundlePath);

    if (virDirOpenIfExists(&dir, dirpath) <= 0)
        goto cleanup;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    while (virDirRead(dir, &entry, NULL) > 0) {
        unsigned int flags = VIR_DOMAIN_SNAPSHOT_PARSE_REDEFINE |
                             VIR_DOMAIN_SNAPSHOT_PARSE_INTERNAL;
        g_autofree char *path = NULL;
        g_autofree char *xml = NULL;
        virDomainSnapshotDefPtr def;
        virDomainMomentObjPtr snap;
        char uuid[VIR_UUID_STRING_BUFLEN + 2];

        if (virStrncpy(uuid, entry->d_name, sizeof(uuid) - 1, sizeof(uuid)) < 0) {
            VIR_WARN("unexpected snapshot file name=%s", entry->d_name);
            continue;
        }

        path = g_strdup_printf("%s/%s", dirpath, entry->d_name);
        if (!virDomainSnapshotFindByName(priv->snapshots, uuid)) {
            if (unlink(path) < 0)
                VIR_WARN("Failed to unlink path=%s, errno=%d", path, errno);
            continue;
        }

        if (!(xml = virFileRead(path))) {
            VIR_WARN("Failed to read snapshot file=%s, errno=%d", path, errno);
            continue;
        }
        if (!(def = virDomainSnapshotDefParseString(xml,
                                                    driver->xmlopt,
                                                    NULL, NULL, flags)))
            continue;

        if (!(snap = virDomainSnapshotAssignDef(dom->snapshots, def)))
            virObjectUnref(def);
    }

 cleanup:
    virDomainSnapshotObjListRemoveAll(priv->snapshots);
    virDirClose(&dir);
    return 0;
}


static int
vzctAutostartDomain(virDomainObjPtr dom,
                    void *opaque)
{
    vzctDriverPtr driver = opaque;

    virObjectLock(dom);

    if (!dom->autostart || virDomainObjIsActive(dom))
        goto cleanup;

    if (vzctDomainStartFull(driver, dom) < 0)
        goto cleanup;

 cleanup:
    virObjectUnlock(dom);
    return 0;
}


static int
vzctStateInitialize(bool privileged,
                    const char *root G_GNUC_UNUSED,
                    virStateInhibitCallback callback G_GNUC_UNUSED,
                    void *opaque G_GNUC_UNUSED)
{
    vzctDriverPtr driver = NULL;
    bool autostart;

    /* Check that the user is root, silently disable if not */
    if (!privileged) {
        VIR_DEBUG("Not running privileged, disabling driver");
        return 0;
    }

    VIR_ALLOC(driver);

    if (virMutexInit(&driver->lock) < 0) {
        VIR_FREE(driver);
        return -1;
    }
    vzct_driver = driver;

    if (!(driver->xmlopt = virDomainXMLOptionNew(&vzctDriverDomainDefParserConfig,
                                                 &vzctDriverPrivateDataCallbacks,
                                                 NULL, NULL, NULL))) {
        goto error;
    }

    VIR_ALLOC(driver->config);
    driver->config->stateDir = g_strdup(LOCALSTATEDIR"/run/libvirt/vzct");

    if (virFileMakePath(driver->config->stateDir) < 0) {
        virReportSystemError(errno, _("Failed to create state dir %s"),
                             driver->config->stateDir);
        goto error;
    }

    if (!(driver->caps = vzctDriverCapsInit()))
        goto error;

    if (!(driver->vncPorts = virPortAllocatorRangeNew(_("vnc"), 5900, 65535)))
        goto error;

    if (!(driver->workers = virThreadPoolNew(0, 1, 0,
                                             vzctDriverJobHandler,
                                             vzct_driver)))
        goto error;

    if (!(driver->eventsManager = virObjectEventStateNew()))
        goto error;

    if (vzctDriverConnectOpen(driver) < 0)
        goto error;


    vzctCleanupDriverConfigs(driver);

    if (virDriverShouldAutostart(driver->config->stateDir, &autostart) < 0)
        goto error;

    if (autostart)
        virDomainObjListForEach(driver->domains, false,
                                vzctAutostartDomain, driver);

    return 0;

 error:
    vzctStateCleanup();
    return -1;
}


static int
vzctConnectSupportsFeature(virConnectPtr conn,
                           int feature)
{
    if (virConnectSupportsFeatureEnsureACL(conn) < 0)
        return -1;

    switch (feature) {
    case VIR_DRV_FEATURE_MIGRATION_PARAMS:
    case VIR_DRV_FEATURE_MIGRATION_P2P:
    case VIR_DRV_FEATURE_MIGRATE_CHANGE_PROTECTION:
    case VIR_DRV_FEATURE_TYPED_PARAM_STRING:
        return 1;
    default:
        return 0;
    }
}


static const char *
vzctConnectGetType(virConnectPtr conn)
{
    if (virConnectGetTypeEnsureACL(conn) < 0)
        return NULL;

    return "vzct";
}


static virDomainObjPtr
vzctDomObjFromDomain(virDomainPtr domain)
{
    vzctDriverPtr driver = domain->conn->privateData;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    virDomainObjPtr dom;

    if (!(dom = virDomainObjListFindByUUID(driver->domains, domain->uuid))) {
        virUUIDFormat(domain->uuid, uuidstr);
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("no domain with matching uuid '%s' (%s)"),
                       uuidstr, domain->name);
        return NULL;
    }

    return dom;
}


static int
vzctConnectListAllDomains(virConnectPtr conn,
                          virDomainPtr **doms,
                          unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    int ret = -1;

    virCheckFlags(VIR_CONNECT_LIST_DOMAINS_FILTERS_ALL, -1);

    if (virConnectListAllDomainsEnsureACL(conn) < 0)
        return -1;

    ret = virDomainObjListExport(driver->domains, conn, doms,
                                 virConnectListAllDomainsCheckACL, flags);

    return ret;
}


static int
vzctDomainGetState(virDomainPtr domain,
                   int *state,
                   int *reason,
                   unsigned int flags)
{
    virDomainObjPtr dom;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetStateEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    *state = virDomainObjGetState(dom, reason);
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static virDomainPtr
vzctDomainLookupByUUID(virConnectPtr conn,
                       const unsigned char *uuid)
{
    vzctDriverPtr driver = conn->privateData;
    virDomainObjPtr dom;
    virDomainPtr domain = NULL;

    if (!(dom = virDomainObjListFindByUUID(driver->domains, uuid))) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];
        virUUIDFormat(uuid, uuidstr);
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching uuid '%s'"), uuidstr);
        goto cleanup;
    }

    if (virDomainLookupByUUIDEnsureACL(conn, dom->def) < 0)
        goto cleanup;

    domain = virGetDomain(conn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    virDomainObjEndAPI(&dom);
    return domain;
}


static virDomainPtr
vzctDomainLookupByName(virConnectPtr conn,
                       const char *name)
{
    vzctDriverPtr driver = conn->privateData;
    virDomainObjPtr dom;
    virDomainPtr domain = NULL;

    if (!(dom = virDomainObjListFindByName(driver->domains, name))) {
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching name '%s'"), name);
        goto cleanup;
    }

    if (virDomainLookupByNameEnsureACL(conn, dom->def) < 0)
        goto cleanup;

    domain = virGetDomain(conn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    virDomainObjEndAPI(&dom);
    return domain;
}


static char *
vzctDomainGetXMLDesc(virDomainPtr domain,
                     unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom;
    char *ret = NULL;
    bool job = false;

    /* Flags checked by virDomainDefFormat */

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetXMLDescEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    ret = virDomainDefFormat((flags & VIR_DOMAIN_XML_INACTIVE) &&
                             dom->newDef ? dom->newDef : dom->def,
                             driver->xmlopt,
                             virDomainDefFormatConvertXMLFlags(flags));

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(vzctl_env_param, vzctl2_free_env_param);

static int
vzctlDomainCreate(virDomainObjPtr dom,
                  virDomainDef *def)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    g_autoptr(vzctl_env_param) env_param = NULL;
    struct vzctl_env_create_param create_param = {
        /* FIXME */
        .config = (char*) "vswap.1024MB",
        .layout = VZCTL_LAYOUT_5,
        .no_root_disk = 1,
    };

    virUUIDFormat(def->uuid, uuidstr);
    create_param.uuid = uuidstr;
    SET_CTID(create_param.ctid, def->extraId);
    create_param.name = def->name;

    VZCT_CHECK_PTR_RET(-1, env_param, vzctl2_alloc_env_param());
    VZCT_CHECK_RET(-1, vzctl2_env_create(env_param, &create_param, 0));

    def->extraId = g_strdup(uuidstr);
    virDomainObjSetState(dom, VIR_DOMAIN_SHUTOFF, -1);

    return 0;
}


static char *
vzctDomainDiskDescription(const virDomainDiskDef *disk)
{
    g_autofree char *ret = NULL;

    switch (disk->src->type) {
    case VIR_STORAGE_TYPE_FILE:
        ret = g_strdup_printf("file='%s'", disk->src->path);
        break;

    case VIR_STORAGE_TYPE_VOLUME:
        ret = g_strdup_printf("pool='%s', volume='%s'",
                              disk->src->srcpool->pool,
                              disk->src->srcpool->volume);
        break;

    default:
        ret = g_strdup("type 'unsupported'");
        break;
    }

    return g_steal_pointer(&ret);
}


static int
vzctConfigDiskAdd(vzctl_env_handle_ptr env,
                  const virDomainDiskDef *disk)
{
    /* just { 0 } is known bug for gcc less then 5.1 */
    struct vzctl_disk_param params = { .enabled = 0 };
    vzctDomainDiskPrivatePtr diskPrivate =
        (vzctDomainDiskPrivatePtr)disk->privateData;
    char rootpath[2] = "/";
    char emptypath[1] = "";
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    unsigned char uuid[VIR_UUID_BUFLEN];

    VIR_DEBUG("Adding disk to config %s", vzctDomainDiskDescription(disk));

    params.enabled = 1;

    switch (disk->src->type) {
    case VIR_STORAGE_TYPE_FILE:
        params.path = disk->src->path;
        break;

    case VIR_STORAGE_TYPE_VOLUME:
        params.path = emptypath;
        break;

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("unsupported disk source type '%s'"),
                       virStorageTypeToString(disk->src->type));
        return -1;
    }

    if (disk->info.bootIndex == 1)
        params.mnt = rootpath;

    if (virUUIDGenerate(uuid) < 0)
        return -1;
    virUUIDFormat(uuid, uuidstr);
    memcpy(params.uuid + 1, uuidstr, VIR_UUID_STRING_BUFLEN);
    params.uuid[0] = '{';
    params.uuid[VIR_UUID_STRING_BUFLEN] = '}';
    memcpy(diskPrivate->uuid, params.uuid, sizeof(params.uuid));

    VZCT_CHECK_RET(-1, vzctl2_env_attach_disk(env, &params));
    return 0;
}


static int
vzctConfigDiskDelete(vzctl_env_handle_ptr env,
                     const virDomainDiskDef *disk)
{
    const vzctDomainDiskPrivate *diskPrivate =
        (const vzctDomainDiskPrivate *)disk->privateData;

    VIR_DEBUG("Deleting disk from config %s", vzctDomainDiskDescription(disk));

    VZCT_CHECK_RET(-1, vzctl2_env_detach_disk(env, diskPrivate->uuid));
    return 0;
}


static int
vzctConfigDiskUpdate(vzctl_env_handle_ptr env,
                     const virDomainDiskDef *olddisk,
                     virDomainDiskDef *disk)
{
    /* just { 0 } is known bug for gcc less then 5.1 */
    struct vzctl_disk_param param = { .enabled = 0 };
    const vzctDomainDiskPrivate *oldDiskPrivate =
        (const vzctDomainDiskPrivate *)olddisk->privateData;
    vzctDomainDiskPrivate *diskPrivate =
        (vzctDomainDiskPrivate *)disk->privateData;
    char rootpath[2] = "/";
    char emptypath[1] = "";

    VIR_DEBUG("Updating disk config '%s'", vzctDomainDiskDescription(disk));

    memcpy(diskPrivate->uuid, oldDiskPrivate->uuid, sizeof(diskPrivate->uuid));

    if (disk->info.bootIndex == olddisk->info.bootIndex)
        return 0;

    memcpy(param.uuid, oldDiskPrivate->uuid, sizeof(param.uuid));

    if (disk->info.bootIndex == 1)
        param.mnt = rootpath;
    else
        param.mnt = emptypath;

    VZCT_CHECK(error,
               vzctl2_env_set_disk(env, &param));

    return 0;

 error:
    return -1;
}


static int
vzctDomainDiskMatch(const virDomainDiskDef *a,
                    const virDomainDiskDef *b)
{
    if (a->src->type != b->src->type)
        return false;

    switch (a->src->type) {
    case VIR_STORAGE_TYPE_FILE:
        return STREQ_NULLABLE(a->src->path, b->src->path);

    case VIR_STORAGE_TYPE_VOLUME:
        return STREQ_NULLABLE(a->src->srcpool->pool,
                              b->src->srcpool->pool) &&
               STREQ_NULLABLE(a->src->srcpool->volume,
                              b->src->srcpool->volume);

    default:
         return false;
    }

    return false;
}


static int
vzctConfigSetDisks(vzctl_env_handle_ptr env,
                   const virDomainDef *olddef,
                   const virDomainDef *def)
{
    size_t i, j;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        const virDomainDiskDef *olddisk = NULL;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        for (j = 0; olddef && j < olddef->ndisks; j++) {
            const virDomainDiskDef *d = olddef->disks[j];

            if (vzctDomainDiskMatch(disk, d)) {
                olddisk = olddef->disks[j];
                break;
            }
        }

        if (olddisk) {
            if (vzctConfigDiskUpdate(env, olddisk, disk) < 0)
                return -1;
        } else {
            if (vzctConfigDiskAdd(env, disk) < 0)
                return -1;
        }
    }

    for (i = 0; olddef && i < olddef->ndisks; i++) {
        const virDomainDiskDef *olddisk = olddef->disks[i];
        const virDomainDiskDef *disk = NULL;

        if (olddisk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        for (j = 0; j < def->ndisks; j++) {
            if (vzctDomainDiskMatch(olddisk, def->disks[j])) {
                disk = def->disks[j];
                break;
            }
        }

        if (!disk && vzctConfigDiskDelete(env, olddisk) < 0)
            return -1;
    }

    return 0;
}


static char *
virNetDevIpAddrFormat(const virNetDevIPAddr *ip)
{
    g_autofree char *iponly = NULL;
    g_autofree char *ipmask = NULL;

    if (!(iponly = virSocketAddrFormat(&ip->address)))
        return NULL;

    if (ip->prefix) {
        if (VIR_SOCKET_ADDR_IS_FAMILY(&ip->address, AF_INET6)) {
            ipmask = g_strdup_printf("%s/%u", iponly, ip->prefix);
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&ip->address, AF_INET)) {
            g_autofree char *maskstr = NULL;
            virSocketAddr mask;

            if (virSocketAddrPrefixToNetmask(ip->prefix, &mask, AF_INET) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("invalid prefix %u"), ip->prefix);
                return NULL;
            }

            if (!(maskstr = virSocketAddrFormat(&mask)))
                return NULL;

            ipmask = g_strdup_printf("%s/%s", iponly, maskstr);
        }
    }

    if (ipmask)
        return g_steal_pointer(&ipmask);
    else
        return g_steal_pointer(&iponly);
}


static int
vzctConfigNetAdd(vzctl_env_param_ptr param,
                 const virDomainNetDef *net,
                 struct vzctl_veth_dev_param *oldnet_param)
{
    struct vzctl_veth_dev_param net_param;
    vzctl_veth_dev_iterator it = NULL;
    char mac[VIR_MAC_STRING_BUFLEN];
    size_t i;

    VIR_DEBUG("Updating/adding network interface: %s", net->ifname);

    if (oldnet_param)
        memcpy(&net_param, oldnet_param, sizeof(net_param));
    else
        memset(&net_param, 0, sizeof(net_param));

    virMacAddrFormat(&net->mac, mac);
    net_param.mac = mac;

    net_param.dev_name = net->ifname;
    net_param.dev_name_ve = net->ifname_guest;
    switch ((int)net->type) {
    case VIR_DOMAIN_NET_TYPE_BRIDGE:
        net_param.nettype = VZCTL_NETTYPE_BRIDGE;
        net_param.network = net->data.bridge.brname;
        break;

    case VIR_DOMAIN_NET_TYPE_NETWORK:
        net_param.nettype = VZCTL_NETTYPE_VNET;
        net_param.network = net->data.network.name;
        break;

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unexpected net interface type: %s"),
                       virDomainNetTypeToString(net->type));
        return -1;
    }
    if (net->virtPortProfile &&
        net->virtPortProfile->virtPortType == VIR_NETDEV_VPORT_PROFILE_OPENVSWITCH)
        net_param.vporttype = VZCTL_VPORTTYPE_OVS;
    else
        net_param.vporttype = VZCTl_VPORTTYPE_NONE;

    VZCT_CHECK_PTR_RET(-1, it, vzctl2_create_veth_dev(&net_param, sizeof(net_param)));

    VZCT_CHECK_RET(-1, vzctl2_env_del_veth_ipaddress(it, "all"));
    for (i = 0; i < net->guestIP.nips; i++) {
        g_autofree char *ipstr = NULL;

        if (!(ipstr = virNetDevIpAddrFormat(net->guestIP.ips[i])))
            return -1;

        VZCT_CHECK_RET(-1, vzctl2_env_add_veth_ipaddress(it, ipstr));
    }

    VZCT_CHECK_RET(-1, vzctl2_env_add_veth(param, it));
    return 0;
}


static int
vzctConfigNetDelete(vzctl_env_param_ptr param,
                    const char *ifname)
{
    VIR_DEBUG("Deleting network interface: %s", ifname);

    VZCT_CHECK_RET(-1, vzctl2_env_del_veth(param, ifname));
    return 0;
}


static int
vzctConfigUpdateVenet0(vzctl_env_param_ptr param,
                       const virDomainNetDef *net)
{
    size_t i;

    VZCT_CHECK_RET(-1, vzctl2_env_del_ipaddress(param, "all"));

    for (i = 0; i < net->guestIP.nips; i++) {
        g_autofree char *ipstr = NULL;

        if (!(ipstr = virNetDevIpAddrFormat(net->guestIP.ips[i])))
            return -1;

        VZCT_CHECK_RET(-1, vzctl2_env_add_ipaddress(param, ipstr));
    }

    return 0;
}


static int
vzctConfigSetNets(vzctl_env_param_ptr param,
                  vzctl_env_param_ptr oldparam,
                  const virDomainDef *def)
{
    size_t i;
    vzctl_veth_dev_iterator it;

    for (i = 0; i < def->nnets; i++) {
        const virDomainNetDef *net = def->nets[i];
        struct vzctl_veth_dev_param oldnet_param;

        if (net->type == VIR_DOMAIN_NET_TYPE_VENET0) {
            if (vzctConfigUpdateVenet0(param, net) < 0)
                return -1;

            continue;
        }

        it = NULL;
        while ((it = vzctl2_env_get_veth(oldparam, it))) {
            vzctl2_env_get_veth_param(it, &oldnet_param, sizeof(oldnet_param));
            if (STREQ_NULLABLE(net->ifname, oldnet_param.dev_name))
                break;
        }

        if (vzctConfigNetAdd(param, net, it ? &oldnet_param : NULL) < 0)
            return -1;
    }

    it = NULL;
    while ((it = vzctl2_env_get_veth(oldparam, it))) {
        struct vzctl_veth_dev_param oldnet_param;
        const virDomainNetDef *net = NULL;

        vzctl2_env_get_veth_param(it, &oldnet_param, sizeof(oldnet_param));
        for (i = 0; i < def->nnets; i++) {
            if (STREQ_NULLABLE(oldnet_param.dev_name, def->nets[i]->ifname)) {
                net = def->nets[i];
                break;
            }
        }
        if (!net && vzctConfigNetDelete(param, oldnet_param.dev_name_ve) < 0)
            return -1;
    }

    return 0;
}


static void
vzctConfigSetIotunes(vzctl_env_param_ptr param,
                     const virDomainDef *def)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        virDomainBlockIoTuneInfoPtr iotune = &disk->blkdeviotune;
        unsigned int io = iotune->total_bytes_sec;
        unsigned int iops = iotune->total_iops_sec;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        /*
         * Quite strange interface. We set 0 here for unlimited but
         * get back UINT_MAX on read.
         */
        vzctl2_env_set_iolimit(param, io);
        vzctl2_env_set_iopslimit(param, iops);
        break;
    }
}


static int
vzctConfigSet(const virDomainDef *olddef,
              const virDomainDef *def)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    struct vzctl_mem_guarantee mem_guarantee;
    vzctl_env_param_ptr oldparam;
    const char *ctid;
    g_autofree char *mask = NULL;
    int err;

    ctid = olddef ? olddef->extraId : def->extraId;
    VZCT_CHECK_PTR_RET(-1, param, vzctl2_alloc_env_param());
    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(ctid, 0, &err));
    oldparam = vzctl2_get_env_param(env);

    /* Ceiling memory size from Kb to Mb */
    VZCT_CHECK_RET(-1, vzctl2_env_set_ramsize(param,
                                              (def->mem.cur_balloon + 1023)
                                              / 1024));
    memset(&mem_guarantee, 0, sizeof(mem_guarantee));
    if (def->mem.min_guarantee_auto) {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_AUTO;
    } else {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_BYTES;
        /* KiB to bytes */
        mem_guarantee.value = def->mem.min_guarantee << 10;
    }
    VZCT_CHECK_RET(-1, vzctl2_env_set_memguarantee(param, &mem_guarantee));

    VZCT_CHECK_RET(-1, vzctl2_env_set_cpu_count(param, def->maxvcpus));
    if (!(mask = virBitmapFormat(def->cpumask)))
        return -1;
    VZCT_CHECK_RET(-1, vzctl2_env_set_cpumask(param, mask));

    VIR_FREE(mask);
    if (!(mask = virDomainNumatuneFormatNodeset(def->numa, NULL, -1)))
        return -1;
    VZCT_CHECK_RET(-1, vzctl2_env_set_nodemask(param, mask));

    if (vzctConfigSetDisks(env, olddef, def) < 0)
        return -1;

    if (vzctConfigSetNets(param, oldparam, def) < 0)
        return -1;

    vzctConfigSetIotunes(param, def);

    VZCT_CHECK_RET(-1, vzctl2_apply_param(env, param, VZCTL_SAVE));
    return 0;
}


static virDomainPtr
vzctDomainDefineXMLFlags(virConnectPtr conn,
                         const char *xml,
                         unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    vzctDomainObjPrivatePtr priv;
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(virDomainDef) newdef = NULL;
    virDomainObjPtr dom = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE;
    virDomainPtr domain = NULL;
    g_autofree char *bundlePath = NULL;
    virObjectEventPtr event;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    const char *ctid;
    bool job = false;
    int detail;

    if (flags & VIR_DOMAIN_DEFINE_VALIDATE)
        parse_flags |= VIR_DOMAIN_DEF_PARSE_VALIDATE_SCHEMA;

    if ((def = virDomainDefParseString(xml, driver->xmlopt,
                                       NULL, parse_flags)) == NULL)
        return NULL;

    if (virDomainDefineXMLFlagsEnsureACL(conn, def) < 0)
        return NULL;

    virUUIDFormat(def->uuid, uuidstr);
    def->extraId = g_strdup(uuidstr);
    if (!(dom = virDomainObjListAdd(driver->domains, def, driver->xmlopt, 0)))
        return NULL;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (dom->def && virDomainObjGetState(dom, NULL) != VIR_DOMAIN_SHUTOFF) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("define is supportend only for shutoff domains"));
        goto cleanup;
    }

    if (!dom->def && vzctlDomainCreate(dom, def) < 0)
        goto cleanup;

    if (vzctConfigSet(dom->def, def) < 0)
        goto cleanup;

    if (dom->def)
        detail = VIR_DOMAIN_EVENT_DEFINED_UPDATED;
    else
        detail = VIR_DOMAIN_EVENT_DEFINED_ADDED;

    priv = dom->privateData;
    ctid = dom->def ? dom->def->extraId : def->extraId;
    if (!(newdef = vzctLoadDomainConfig(driver, ctid, &bundlePath)))
        goto cleanup;
    virDomainDefFree(dom->def);
    vzctMergeConfigs(newdef, def);
    dom->def = g_steal_pointer(&newdef);
    if (!priv->bundlePath)
        priv->bundlePath = g_steal_pointer(&bundlePath);
    dom->persistent = true;

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     detail);
    virObjectEventStateQueue(driver->eventsManager, event);
    domain = virGetDomain(conn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    if (!domain && !dom->def) {
        /* Removing needs def for name and uuid */
        dom->def = g_steal_pointer(&def);
        virDomainObjListRemove(driver->domains, dom);
    }

    virDomainObjEndAPI(&dom);
    return domain;
}


static virDomainPtr
vzctDomainDefineXML(virConnectPtr conn, const char *xml)
{
    return vzctDomainDefineXMLFlags(conn, xml, 0);
}


static char *
vzctConnectGetCapabilities(virConnectPtr conn)
{
    vzctDriverPtr driver = conn->privateData;
    char *xml;

    if (virConnectGetCapabilitiesEnsureACL(conn) < 0)
        return NULL;

    xml = virCapabilitiesFormatXML(driver->caps);
    return xml;
}


static int
vzctDomainUndefineFlags(virDomainPtr domain,
                        unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom;
    g_autoptr(vzctl_env_handle) env = NULL;
    virObjectEventPtr event;
    size_t i;
    int ret = -1;
    int err;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainUndefineFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("cannot undefine running domain"));
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env,
                   vzctl2_env_open(dom->def->extraId, 0, &err));

    for (i = 0; i < dom->def->ndisks; i++) {
        virDomainDiskDefPtr disk = dom->def->disks[i];
        vzctDomainDiskPrivatePtr diskPrivate =
            (vzctDomainDiskPrivatePtr)disk->privateData;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        /*
         * Disk are detached because vzctl2_env_destroy unlinks disks
         * as well. But in case of volumes we don't store path to disks
         * in offline configuration. And currently undefine works only
         * for offline containers.
         */
        if (disk->src->type == VIR_STORAGE_TYPE_VOLUME)
            continue;

        VZCT_CHECK(cleanup, vzctl2_env_detach_disk(env, diskPrivate->uuid));
    }

    VZCT_CHECK(cleanup, vzctl2_env_destroy(env, 0));
    virDomainObjListRemove(driver->domains, dom);
    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_UNDEFINED,
                                     VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainUndefine(virDomainPtr domain)
{
    return vzctDomainUndefineFlags(domain, 0);
}


static int
vzctDomainIsActive(virDomainPtr domain)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom;
    int ret = -1;
    bool job = false;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainIsActiveEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    ret = virDomainObjIsActive(dom);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainStopVzctl(virDomainObjPtr dom,
                    bool gracefull)
{
    int mode = gracefull ? M_HALT : M_KILL;
    g_autoptr(vzctl_env_handle) env = NULL;
    int ret = -1;
    int err;

    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_RET(-1, vzctl2_env_stop(env, mode, 0));
    ret = 0;
    return ret;
}


static int
vzctDomainAddCdrom(vzctl_env_handle_ptr env,
                   const virDomainDef *def)
{
    vzctl_env_param_ptr param = vzctl2_get_env_param(env);
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *disk = def->disks[i];

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM) {
            VZCT_CHECK_RET(-1, vzctl2_env_set_cidata_fname(param, disk->src->path));
            break;
        }
    }

    return 0;
}


static int
vzctDomainCleanupGarbage(virDomainObjPtr dom)
{
    g_autofree char *migfile = vzctMigFile(dom);

    if (unlink(migfile) < 0 && errno != ENOENT) {
        virReportSystemError(errno,
                             _("cannot remove migrating status file path=%s"),
                             migfile);
        return -1;
    }

    return 0;
}


static int
vzctDomainResolveVolumes(virDomainDefPtr def,
                         vzctl_env_handle *env)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        struct vzctl_disk_param param = { .enabled = 0 };
        const vzctDomainDiskPrivate *diskPrivate =
            (const vzctDomainDiskPrivate *)disk->privateData;

        if (virDomainDiskTranslateSourcePool(disk) < 0)
            return -1;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        if (disk->src->srcpool->actualtype != VIR_STORAGE_TYPE_FILE) {
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                           _("only ploop volumes are supported"));
            return -1;
        }

        memcpy(param.uuid, diskPrivate->uuid, sizeof(param.uuid));
        param.path = disk->src->path;
        VZCT_CHECK_RET(-1, vzctl2_env_set_disk(env, &param));
    }

    return 0;
}


static void
vzctDomainClearVolumes(virDomainDefPtr def,
                       vzctl_env_handle *env)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        struct vzctl_disk_param param = { .enabled = 0 };
        const vzctDomainDiskPrivate *diskPrivate =
            (const vzctDomainDiskPrivate *)disk->privateData;
        char emptypath[1] = "";
        int rc;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        memcpy(param.uuid, diskPrivate->uuid, sizeof(param.uuid));
        param.path = emptypath;

        if ((rc = vzctl2_env_set_disk(env, &param)))
            VZCT_ERROR_MESSAGE_CODE(vzctl2_env_set_disk, rc);
    }
}


static int
vzctDomainStartFull(vzctDriverPtr driver,
                    virDomainObjPtr dom)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    virObjectEventPtr event;
    bool started = false;
    bool job = false;
    int ret = -1;
    int err;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is already running"));
        goto cleanup;
    }

    if (vzctDomainCleanupGarbage(dom) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (vzctDomainAddCdrom(env, dom->def) < 0)
        goto cleanup;

    if (vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_env_start(env, 0));
    started = true;
    if (vzctDomainStart(driver, dom, NULL) < 0)
        goto cleanup;

    event = virDomainEventLifecycleNewFromObj(dom,
                                 VIR_DOMAIN_EVENT_STARTED,
                                 VIR_DOMAIN_EVENT_STARTED_BOOTED);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    if (ret < 0 && started)
        vzctDomainStopVzctl(dom, true);
    if (ret < 0 && env)
        vzctDomainClearVolumes(dom->def, env);
    if (job)
        vzctDomainEndJob(dom);

    return ret;
}


static int
vzctDomainCreateWithFlags(virDomainPtr domain,
                          unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCreateWithFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainStartFull(driver, dom) < 0)
        goto cleanup;

    ret = 0;

 cleanup:

    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCreate(virDomainPtr domain)
{
    return vzctDomainCreateWithFlags(domain, 0);
}


static int
vzctDomainDestroyFlags(virDomainPtr domain,
                       unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virObjectEventPtr event;
    virDomainObjPtr dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainDestroyFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (vzctDomainStopVzctl(dom, false) < 0)
        goto cleanup;

    vzctDomainStop(driver, dom);
    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_DESTROYED);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainDestroy(virDomainPtr domain)
{
    return vzctDomainDestroyFlags(domain, 0);
}


static int
vzctDomainShutdownFlags(virDomainPtr domain,
                        unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virObjectEventPtr event;
    virDomainObjPtr dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainShutdownFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (vzctDomainStopVzctl(dom, true) < 0)
        goto cleanup;

    vzctDomainStop(driver, dom);
    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_SHUTDOWN);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainShutdown(virDomainPtr domain)
{
    return vzctDomainShutdownFlags(domain, 0);
}


static int
vzctDomainGetInfo(virDomainPtr domain,
                  virDomainInfoPtr info)
{
    vzctDriverPtr driver = domain->conn->privateData;
    struct vzctl_cpustat cpustat;
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainObjPtr dom;
    bool job = false;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetInfoEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    info->state = virDomainObjGetState(dom, NULL);
    info->memory = dom->def->mem.cur_balloon;
    info->maxMem = virDomainDefGetMemoryTotal(dom->def);
    info->nrVirtCpu = virDomainDefGetVcpusMax(dom->def);


    if (virDomainObjIsActive(dom)) {
        int err;

        VZCT_CHECK_PTR(cleanup, env,
                       vzctl2_env_open(dom->def->extraId,
                                       VZCTL_CONF_SKIP_PARSE, &err));

        VZCT_CHECK(cleanup, vzctl2_env_cpustat(env, &cpustat, sizeof(cpustat)));
        /* secs to nanosecs */
        info->cpuTime = (cpustat.user + cpustat.system + cpustat.nice) * 1000000000ULL;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainIsPersistent(virDomainPtr domain)
{
    virDomainObjPtr dom;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainIsPersistentEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = 1;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static char *
vzctDomainGetOSType(virDomainPtr domain)
{
    virDomainObjPtr dom;
    char *ret = NULL;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetOSTypeEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = g_strdup(virDomainOSTypeToString(dom->def->os.type));

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCollectNetStats(virTypedParameterPtr *params,
                          int *nparams,
                          int *maxparams,
                          vzctl_env_handle_ptr env,
                          virDomainNetDefPtr net,
                          size_t idx)
{
    const char *ctid = vzctl2_env_get_ctid(env);
    char name[VIR_TYPED_PARAM_FIELD_LENGTH];
    struct vzctl_netstat stats;
    const char *ifname;

    if (net->type == VIR_DOMAIN_NET_TYPE_VENET0)
        ifname = VZCT_NET_VENET0;
    else
        ifname = net->ifname;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.name", idx);
    if (virTypedParamsAddString(params, nparams, maxparams, name,
                                ifname) < 0)
        return -1;

    if (!env)
        return 0;

    VZCT_CHECK(error,
               vzctl2_get_env_netstat(ctid, ifname, &stats, sizeof(stats)));

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.rx.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.incoming) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.rx.pkts", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.incoming_pkt) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.tx.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.outgoing) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.tx.pkts", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.outgoing_pkt) < 0)
        return -1;

    return 0;

 error:
    return -1;
}


static int
vzctDomainCollectDiskStats(virTypedParameterPtr *params,
                           int *nparams,
                           int *maxparams,
                           vzctl_env_handle_ptr env,
                           virDomainDiskDefPtr disk,
                           size_t idx)
{
    vzctDomainDiskPrivatePtr priv = (vzctDomainDiskPrivatePtr)disk->privateData;
    char name[VIR_TYPED_PARAM_FIELD_LENGTH];
    struct vzctl_disk_stats stats;
    unsigned long long usage;
    int rc;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.name", idx);
    if (virTypedParamsAddString(params, nparams, maxparams, name,
                                disk->dst) < 0)
        return -1;

    /*
     * In case of offline domain volume path etc can be outdated. So don't
     * report anything besides name. This is on par with qemu driver.
     */
    if (disk->src->type == VIR_STORAGE_TYPE_VOLUME && !env)
        return 0;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.path", idx);
    if (virTypedParamsAddString(params, nparams, maxparams, name,
                                disk->src->path) < 0)
        return -1;

    VZCT_CHECK_RET(-1, vzctl2_get_disk_usage(disk->src->path, &usage));

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.physical", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name, usage) < 0)
        return -1;

    if (!env)
        return 0;

    rc = vzctl2_env_get_disk_stats(env, priv->uuid, &stats, sizeof(stats));
    if (rc == VZCTL_E_NOTSUPP) {
        /* disk is attached backup probably, stats are not provided */
        return 0;
    } else if (rc) {
        VZCT_ERROR_MESSAGE_CODE(vzctl2_env_get_disk_stats, rc);
        return -1;
    }

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.rd.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.read) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.wr.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.write) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.rd.reqs", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.read_reqs) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.wr.reqs", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.write_reqs) < 0)
        return -1;
    /* convert kB to bytes */
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.capacity", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.total << 10) < 0)
        return -1;

    return 0;
}


static virDomainStatsRecordPtr
vzctDomainCollectStats(virConnectPtr conn,
                       virDomainObjPtr dom,
                       int stats,
                       unsigned int flags)
{
    virDomainDefPtr def = dom->def;
    virDomainStatsRecordPtr tmp;
    virDomainStatsRecordPtr ret = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    int maxparams = 0;
    long long timeout;
    bool job = false;
    size_t i;
    int err;

    virObjectLock(dom);
    VIR_ALLOC(tmp);

    if (flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_NOWAIT)
        timeout = 0;
    else
        timeout = VZCT_JOB_WAIT_TIME;

    if (vzctDomainBeginJobInternal(dom, VZCT_DOMAIN_JOB_QUERY, timeout) < 0)
        goto cleanup;
    job = true;

    if (dom->state.state == VIR_DOMAIN_NOSTATE)
        stats &= VIR_DOMAIN_STATS_STATE;

    if (virDomainObjIsActive(dom))
        VZCT_CHECK_PTR(cleanup, env,
                       vzctl2_env_open(dom->def->extraId, 0, &err));

    if (stats & VIR_DOMAIN_STATS_STATE) {
        int state;
        int reason;

        state = virDomainObjGetState(dom, &reason);
        if (virTypedParamsAddInt(&tmp->params, &tmp->nparams, &maxparams,
                                 "state.state", state) < 0 ||
            virTypedParamsAddInt(&tmp->params, &tmp->nparams, &maxparams,
                                 "state.reason", reason) < 0)
            goto cleanup;
    }

    if ((stats & VIR_DOMAIN_STATS_CPU_TOTAL) && env) {
        struct vzctl_cpustat cpustats;
        unsigned long long time;

        VZCT_CHECK(cleanup, vzctl2_env_cpustat(env, &cpustats, sizeof(cpustats)));

        /* secs to nanosecs */
        time = (cpustats.user + cpustats.nice + cpustats.system) * 1000000000ULL;
        if (virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "cpu.time", time) < 0 ||
            virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "cpu.user",
                                    cpustats.user * 1000000000ULL) < 0 ||
            virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "cpu.system",
                                    cpustats.system * 1000000000ULL) < 0)
            goto cleanup;
    }

    if (stats & VIR_DOMAIN_STATS_BALLOON) {
        unsigned long long mem = virDomainDefGetMemoryTotal(def);

        if (virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "balloon.current", mem) < 0 ||
            virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "balloon.maximum", mem) < 0)
            goto cleanup;

        if (virDomainObjIsActive(dom)) {
            struct vzctl_meminfo meminfo;

            VZCT_CHECK(cleanup,
                       vzctl2_get_env_meminfo(dom->def->extraId,
                                              &meminfo, sizeof(meminfo)));

            /* bytes to KiB */
            if (virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.available",
                                        meminfo.total >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.unused",
                                        meminfo.free >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.disk_caches",
                                        meminfo.cached >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.swap_in",
                                        meminfo.swap_in >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.swap_out",
                                        meminfo.swap_out >> 10) < 0)
            goto cleanup;
        }
    }

    if (stats & VIR_DOMAIN_STATS_VCPU) {
        int nvcpus = def->maxvcpus;

        /* unlimited */
        if (!nvcpus && (nvcpus = virHostCPUGetCount()) < 0)
            goto cleanup;

        if (virTypedParamsAddUInt(&tmp->params, &tmp->nparams, &maxparams,
                                  "vcpu.current", nvcpus) < 0 ||
            virTypedParamsAddUInt(&tmp->params, &tmp->nparams, &maxparams,
                                  "vcpu.maximum", nvcpus) < 0)
            goto cleanup;
    }

    if (stats & VIR_DOMAIN_STATS_INTERFACE) {
        if (virTypedParamsAddUInt(&tmp->params, &tmp->nparams,
                                  &maxparams, "net.count", def->nnets) < 0)
            goto cleanup;

        for (i = 0; i < def->nnets; i++) {
            if (vzctDomainCollectNetStats(&tmp->params,
                                          &tmp->nparams, &maxparams,
                                          env, def->nets[i], i) < 0)
                goto cleanup;
        }
    }

    if (stats & VIR_DOMAIN_STATS_BLOCK) {
        size_t ndisks = 0;
        size_t idx = 0;

        for (i = 0; i < def->ndisks; i++) {
            if (def->disks[i]->device == VIR_DOMAIN_DISK_DEVICE_DISK)
                ndisks++;
        }

        if (virTypedParamsAddUInt(&tmp->params, &tmp->nparams,
                                  &maxparams, "block.count", ndisks) < 0)
            goto cleanup;

        for (i = 0; i < def->ndisks; i++) {
            virDomainDiskDefPtr disk = def->disks[i];

            if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
                continue;

            if (vzctDomainCollectDiskStats(&tmp->params,
                                           &tmp->nparams, &maxparams,
                                           env, disk, idx++) < 0)
                goto cleanup;
        }
    }

    if (!(tmp->dom = virGetDomain(conn, def->name, def->uuid, def->id)))
        goto cleanup;

    ret = g_steal_pointer(&tmp);

 cleanup:
    if (tmp) {
        virTypedParamsFree(tmp->params, tmp->nparams);
        VIR_FREE(tmp);
    }
    if (job)
        vzctDomainEndJob(dom);
    virObjectUnlock(dom);
    return ret;
}


static int
vzctConnectGetAllDomainStats(virConnectPtr conn,
                             virDomainPtr *domains,
                             unsigned int ndomains,
                             unsigned int stats,
                             virDomainStatsRecordPtr **retStats,
                             unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    unsigned int supported = VIR_DOMAIN_STATS_STATE |
                             VIR_DOMAIN_STATS_CPU_TOTAL |
                             VIR_DOMAIN_STATS_BALLOON |
                             VIR_DOMAIN_STATS_VCPU |
                             VIR_DOMAIN_STATS_INTERFACE |
                             VIR_DOMAIN_STATS_BLOCK;
    unsigned int lflags = flags & (VIR_CONNECT_LIST_DOMAINS_FILTERS_ACTIVE |
                                   VIR_CONNECT_LIST_DOMAINS_FILTERS_PERSISTENT |
                                   VIR_CONNECT_LIST_DOMAINS_FILTERS_STATE);
    virDomainObjPtr *doms = NULL;
    size_t ndoms;
    size_t i;
    int ret = -1;

    virCheckFlags(VIR_CONNECT_LIST_DOMAINS_FILTERS_ACTIVE |
                  VIR_CONNECT_LIST_DOMAINS_FILTERS_PERSISTENT |
                  VIR_CONNECT_LIST_DOMAINS_FILTERS_STATE |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_NOWAIT |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_ENFORCE_STATS, -1);

    if (virConnectGetAllDomainStatsEnsureACL(conn) < 0)
        return -1;

    if (!stats) {
        stats = supported;
    } else if ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_ENFORCE_STATS) &&
        !(stats & supported)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED,
                       _("Stats types bits 0x%x are not supported"),
                       stats & ~supported);
        return -1;
    }

    if (ndomains) {
        if (virDomainObjListConvert(driver->domains, conn,
                                    domains, ndomains, &doms, &ndoms,
                                    virConnectGetAllDomainStatsCheckACL,
                                    lflags, true) < 0)
            return -1;
    } else {
        if (virDomainObjListCollect(driver->domains, conn, &doms, &ndoms,
                                    virConnectGetAllDomainStatsCheckACL,
                                    lflags) < 0)
            return -1;
    }

    VIR_ALLOC_N(*retStats, ndoms + 1);

    ret = 0;
    for (i = 0; i < ndoms; i++) {
        if (!((*retStats)[ret] = vzctDomainCollectStats(conn, doms[i],
                                                        stats, flags))) {
            virResetLastError();
        } else {
            ret++;
        }
    }

    virObjectListFreeCount(doms, ndoms);

    return ret;
}


static int
vzctDomainHasManagedSaveImage(virDomainPtr domain, unsigned int flags)
{
    virDomainObjPtr dom = NULL;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainHasManagedSaveImageEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSnapshotObjListStealMetaIter(void *payload,
                                       const void *name G_GNUC_UNUSED,
                                       void *data)
{
    virDomainMomentObjPtr snap = payload;
    virDomainMomentObjPtr peer;
    virDomainSnapshotObjListPtr src = data;

    if ((peer = virDomainSnapshotFindByName(src, snap->def->name))) {
        virDomainSnapshotDefPtr src_def = virDomainSnapshotObjGetDef(peer);
        virDomainSnapshotDefPtr dst_def = virDomainSnapshotObjGetDef(snap);

        dst_def->state = src_def->state;
        dst_def->memory = src_def->memory;
        if (!snap->def->dom)
            snap->def->dom = g_steal_pointer(&peer->def->dom);
        if (!snap->def->inactiveDom)
            snap->def->inactiveDom = g_steal_pointer(&peer->def->inactiveDom);
    }

    return 0;
}


static void
vzctDomainSnapshotObjListStealMeta(virDomainSnapshotObjListPtr dst,
                                   virDomainSnapshotObjListPtr src)
{
    virDomainSnapshotForEach(dst, vzctDomainSnapshotObjListStealMetaIter, src);
}


static void
vzctDomainFinishSnapshots(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv;

    if (!dom)
        return;

    priv  = dom->privateData;
    vzctDomainSnapshotObjListStealMeta(dom->snapshots, priv->snapshots);
    virDomainSnapshotObjListRemoveAll(priv->snapshots);
}


static int
vzctDomainReadSnapshots(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    struct vzctl_snapshot_tree* tree = NULL;
    g_autofree char *path = NULL;
    virDomainMomentObjPtr current = NULL;
    size_t i;
    int ret = -1;
    int err;

    path = g_strdup_printf("%s/Snapshots.xml", priv->bundlePath);
    if (!virFileExists(path))
        return 0;

    VZCT_CHECK_PTR(cleanup, tree,
                   vzctl2_open_snapshot_tree(path, &err));

    for (i = 0; i < tree->nsnapshots; i++) {
        struct vzctl_snapshot_data *data = tree->snapshots[i];
        virDomainSnapshotDefPtr def = NULL;
        virDomainMomentObjPtr snapshot;
        struct tm tm;
        const char *tmp;

        if (!(def = virDomainSnapshotDefNew()))
            goto cleanup;

        def->parent.name = g_strdup(data->guid);
        if (data->parent_guid && data->parent_guid[0] != '\0')
            def->parent.parent_name = g_strdup(data->parent_guid);
        def->parent.description = g_strdup(data->desc);
        tmp = strptime(data->date, "%Y-%m-%d %H:%M:%S", &tm);
        if (!tmp || *tmp != '\0') {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("unexpected datetime format: '%s'"), data->date);
            virObjectUnref(def);
            goto cleanup;
        }
        def->parent.creationTime = mktime(&tm);
        /* libvzctl does not provide state and thus memory */
        def->state = VIR_DOMAIN_SNAPSHOT_NOSTATE;
        def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_DEFAULT;

        if (!(snapshot = virDomainSnapshotAssignDef(priv->snapshots, def))) {
            virObjectUnref(def);
            goto cleanup;
        }

        if (data->current) {
            if (current) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("more then one current snapshots"));
                goto cleanup;
            }
            current = snapshot;
        }
    }

    virDomainSnapshotSetCurrent(priv->snapshots, current);
    if (virDomainSnapshotUpdateRelations(priv->snapshots) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("snapshots have inconsistent relations"));
        goto cleanup;
    }

    vzctDomainSnapshotObjListStealMeta(priv->snapshots, dom->snapshots);
    ret = 0;

 cleanup:
    if (ret < 0)
        virDomainSnapshotObjListRemoveAll(priv->snapshots);

    if (tree)
        vzctl2_close_snapshot_tree(tree);

    return ret;
}


static virDomainObjPtr
vzctDomObjFromSnapshot(virDomainSnapshotPtr snapshot)
{
    return vzctDomObjFromDomain(snapshot->domain);
}


static virDomainMomentObjPtr
vzctSnapObjFromName(virDomainSnapshotObjListPtr snapshots,
                    const char *name)
{
    virDomainMomentObjPtr snap = virDomainSnapshotFindByName(snapshots, name);

    if (!(snap = virDomainSnapshotFindByName(snapshots, name)))
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT,
                       _("no domain snapshot with matching name '%s'"),
                       name);

    return snap;
}


static virDomainMomentObjPtr
vzctSnapObjFromSnapshot(virDomainSnapshotObjListPtr snapshots,
                        virDomainSnapshotPtr snapshot)
{
    return vzctSnapObjFromName(snapshots, snapshot->name);
}


static char *
vzctDomainSnapshotGetXMLDesc(virDomainSnapshotPtr snapshot,
                             unsigned int flags)
{
    vzctDriverPtr driver = snapshot->domain->conn->privateData;
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    char *xml = NULL;
    virDomainMomentObjPtr snap = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    bool job = false;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_XML_SECURE, NULL);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return NULL;

    if (virDomainSnapshotGetXMLDescEnsureACL(snapshot->domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    virUUIDFormat(snapshot->domain->uuid, uuidstr);

    xml = virDomainSnapshotDefFormat(uuidstr, virDomainSnapshotObjGetDef(snap),
                                     driver->xmlopt,
                                     virDomainSnapshotFormatConvertXMLFlags(flags));

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return xml;
}


static int
vzctDomainSnapshotListNames(virDomainPtr domain,
                            char **names,
                            int nameslen,
                            unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSnapshotListNamesEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    n = virDomainSnapshotObjListGetNames(priv->snapshots, NULL, names, nameslen,
                                         flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotNum(virDomainPtr domain,
                      unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSnapshotNumEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    n = virDomainSnapshotObjListNum(priv->snapshots, NULL, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainListAllSnapshots(virDomainPtr domain,
                           virDomainSnapshotPtr **snaps,
                           unsigned int flags)
{
    virDomainObjPtr dom = NULL;
    vzctDomainObjPrivatePtr priv;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainListAllSnapshotsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    n = virDomainListSnapshots(priv->snapshots, NULL, domain, snaps, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotNumChildren(virDomainSnapshotPtr snapshot,
                              unsigned int flags)
{
    virDomainObjPtr dom = NULL;
    vzctDomainObjPrivatePtr priv;
    virDomainMomentObjPtr snap = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotNumChildrenEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    n = virDomainSnapshotObjListNum(priv->snapshots, snap, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotListChildrenNames(virDomainSnapshotPtr snapshot,
                                    char **names,
                                    int nameslen,
                                    unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    virDomainMomentObjPtr snap = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotListChildrenNamesEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    n = virDomainSnapshotObjListGetNames(priv->snapshots, snap, names, nameslen,
                                         flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotListAllChildren(virDomainSnapshotPtr snapshot,
                                  virDomainSnapshotPtr **snaps,
                                  unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    virDomainMomentObjPtr snap = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotListAllChildrenEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    n = virDomainListSnapshots(priv->snapshots, snap, snapshot->domain, snaps,
                               flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static virDomainSnapshotPtr
vzctDomainSnapshotLookupByName(virDomainPtr domain,
                               const char *name,
                               unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom;
    virDomainMomentObjPtr snap = NULL;
    virDomainSnapshotPtr snapshot = NULL;
    bool job = false;

    virCheckFlags(0, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotLookupByNameEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromName(priv->snapshots, name)))
        goto cleanup;

    snapshot = virGetDomainSnapshot(domain, snap->def->name);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return snapshot;
}


static int
vzctDomainHasCurrentSnapshot(virDomainPtr domain,
                             unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainHasCurrentSnapshotEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    ret = (virDomainSnapshotGetCurrent(priv->snapshots) != NULL);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static virDomainSnapshotPtr
vzctDomainSnapshotGetParent(virDomainSnapshotPtr snapshot,
                            unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom;
    virDomainMomentObjPtr snap = NULL;
    virDomainSnapshotPtr parent = NULL;
    bool job = false;

    virCheckFlags(0, NULL);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return NULL;

    if (virDomainSnapshotGetParentEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    if (!snap->def->parent_name) {
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT,
                       _("snapshot '%s' does not have a parent"),
                       snap->def->name);
        goto cleanup;
    }

    parent = virGetDomainSnapshot(snapshot->domain, snap->def->parent_name);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return parent;
}


static virDomainSnapshotPtr
vzctDomainSnapshotCurrent(virDomainPtr domain,
                          unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom;
    virDomainSnapshotPtr snapshot = NULL;
    const char *name;
    bool job = false;

    virCheckFlags(0, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotCurrentEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(name = virDomainSnapshotGetCurrentName(priv->snapshots))) {
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT, "%s",
                       _("the domain does not have a current snapshot"));
        goto cleanup;
    }

    snapshot = virGetDomainSnapshot(domain, name);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return snapshot;
}


static int
vzctDomainSnapshotIsCurrent(virDomainSnapshotPtr snapshot,
                            unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    int ret = -1;
    virDomainMomentObjPtr snap = NULL;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotIsCurrentEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    ret = snap == virDomainSnapshotGetCurrent(priv->snapshots);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSnapshotHasMetadata(virDomainSnapshotPtr snapshot,
                              unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    int ret = -1;
    virDomainMomentObjPtr snap = NULL;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotHasMetadataEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    ret = 1;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


/*
 * We don't store `parent` and `is current` values in metadata for
 * the sake on simplicity. These values are taken from libvzctl.
 */
static int
vzctDomainSnapshotWriteMetadata(vzctDriverPtr driver,
                                virDomainObjPtr dom,
                                virDomainMomentObjPtr snapshot)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    virDomainSnapshotDefPtr def = virDomainSnapshotObjGetDef(snapshot);
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    g_autofree char *xml = NULL;
    g_autofree char *dir = NULL;
    g_autofree char *file = NULL;
    unsigned int flags = VIR_DOMAIN_SNAPSHOT_FORMAT_SECURE |
                         VIR_DOMAIN_SNAPSHOT_FORMAT_INTERNAL;

    virUUIDFormat(dom->def->uuid, uuidstr);
    if (!(xml = virDomainSnapshotDefFormat(uuidstr, def,
                                           driver->xmlopt, flags)))
        return -1;

    dir = g_strdup_printf("%s/"VZCT_SNAPHOTS_DIR, priv->bundlePath);
    if (virFileMakePath(dir) < 0) {
        virReportSystemError(errno, _("cannot create directory '%s'"), dir);
        return -1;
    }

    file = g_strdup_printf("%s/%s.xml", dir, def->parent.name);
    return virXMLSaveFile(file, NULL, "snapshot-edit", xml);
}


static virDomainSnapshotPtr
vzctDomainSnapshotCreateXML(virDomainPtr domain,
                            const char *xmlDesc,
                            unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainSnapshotPtr ret = NULL;
    g_autoptr(virDomainSnapshotDef) def = NULL;
    virDomainMomentObjPtr snap = NULL;
    virDomainObjPtr dom = NULL;
    struct vzctl_snapshot_param param = { 0 };
    g_autoptr(vzctl_env_handle) env = NULL;
    unsigned int parse_flags = 0;
    unsigned char uuid[VIR_UUID_BUFLEN];
    char uuidstr[VIR_UUID_STRING_BUFLEN + 2];
    bool job = false;
    int err;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_CREATE_VALIDATE, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotCreateXMLEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (flags & VIR_DOMAIN_SNAPSHOT_CREATE_VALIDATE)
        parse_flags |= VIR_DOMAIN_SNAPSHOT_PARSE_VALIDATE;
    if (!virDomainObjIsActive(dom))
        parse_flags |= VIR_DOMAIN_SNAPSHOT_PARSE_OFFLINE;

    if (!(def = virDomainSnapshotDefParseString(xmlDesc, driver->xmlopt,
                                                NULL, NULL, parse_flags)))
        goto cleanup;

    if (def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_EXTERNAL) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("external memory snapshots are not supported"));
        goto cleanup;
    }

    if (def->parent.name && !def->parent.name_generated) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("can not set name, it is currently only generated"));
        goto cleanup;
    }

    if (virDomainObjIsActive(dom)) {
        if (def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_DEFAULT)
            def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_INTERNAL;

        if (def->memory != VIR_DOMAIN_SNAPSHOT_LOCATION_INTERNAL) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("only internal memory location is supported"
                             " for active domains"));
            goto cleanup;
        }
    } else {
        if (def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_DEFAULT)
            def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_NONE;

        if (def->memory != VIR_DOMAIN_SNAPSHOT_LOCATION_NONE) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("memory location for inactive domains"
                             " must be none"));
            goto cleanup;
        }
    }

    if (!(def->parent.dom = vzctDomainDefCopy(dom->def, driver->xmlopt, true)))
        goto cleanup;

    if (virDomainObjIsActive(dom) &&
        !(def->parent.inactiveDom = vzctDomainDefCopy(dom->newDef,
                                                      driver->xmlopt, true)))
        goto cleanup;

    def->state = virDomainObjGetState(dom, NULL);

    /* libvzctl requires '{uuid}' form of uuid. */
    virUUIDGenerate(uuid);
    virUUIDFormat(uuid, uuidstr + 1);
    uuidstr[0] = '{';
    uuidstr[VIR_UUID_STRING_BUFLEN] = '}';
    uuidstr[VIR_UUID_STRING_BUFLEN + 1] = '\0';
    VIR_FREE(def->parent.name);
    def->parent.name_generated = false;
    def->parent.name = g_strdup(uuidstr);

    if (!(snap = virDomainSnapshotAssignDef(dom->snapshots, def)))
        goto cleanup;
    def = NULL;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (!virDomainObjIsActive(dom) &&
        vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    param.desc = snap->def->description;
    param.guid = snap->def->name;
    VZCT_CHECK_LOCKED(cleanup, env, "creating snapshot",
                      vzctl2_env_create_snapshot(env, &param));

    if (vzctDomainSnapshotWriteMetadata(driver, dom, snap) < 0)
        goto cleanup;

    ret = virGetDomainSnapshot(domain, snap->def->name);

 cleanup:
    if (!ret && snap)
        virDomainSnapshotObjListRemove(dom->snapshots, snap);

    if (env && !virDomainObjIsActive(dom))
        vzctDomainClearVolumes(dom->def, env);

    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainRevertToSnapshot(virDomainSnapshotPtr snapshot,
                           unsigned int flags)
{
    vzctDriverPtr driver = snapshot->domain->conn->privateData;
    virDomainMomentObjPtr snap = NULL;
    virDomainSnapshotDefPtr snapDef;
    virDomainDefPtr def;
    g_autoptr(virDomainDef) drvDef = NULL;
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    struct vzctl_switch_snapshot_param param = { 0 };
    g_autoptr(vzctl_env_handle) env = NULL;
    vzctl_env_status_t status;
    virDomainState state;
    bool job = false;
    int ret = -1;
    int err;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainRevertToSnapshotEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (!virDomainObjIsActive(dom) &&
        vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    param.guid = snap->def->name;
    VZCT_CHECK_LOCKED(cleanup,
                      env, "switching snapshot",
                      vzctl2_env_switch_snapshot(env, &param));

    if (virDomainObjIsActive(dom))
        vzctDomainStop(driver, dom);

    snapDef = virDomainSnapshotObjGetDef(snap);
    /*
     * If snapshot is done thru vzctl than we don't know domain state
     * in snapshot.
     */
    if (snapDef->state == VIR_DOMAIN_NOSTATE) {
        VZCT_CHECK(cleanup,
                   vzctl2_get_env_status(dom->def->extraId, &status,
                                         ENV_STATUS_RUNNING));

        if (status.mask & ENV_STATUS_RUNNING)
            state = VIR_DOMAIN_RUNNING;
        else
            state = VIR_DOMAIN_SHUTOFF;
    } else {
        state = snapDef->state;
    }

    if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
        goto cleanup;
    VIR_FREE(dom->def);
    VIR_FREE(dom->newDef);
    dom->def = g_steal_pointer(&def);
    dom->def->id = -1;

    if (snapDef->parent.inactiveDom)
        drvDef = snapDef->parent.inactiveDom;
    else
        drvDef = snapDef->parent.dom;

    /* Make a copy as merge damages second argument. */
    if (drvDef &&
        !(drvDef = vzctDomainDefCopy(drvDef, driver->xmlopt, false)))
        goto cleanup;
    if (vzctMergeConfigs(dom->def, drvDef) &&
        vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (state == VIR_DOMAIN_RUNNING &&
        vzctDomainStart(driver, dom, snapDef->parent.dom) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    if (env && !virDomainObjIsActive(dom))
        vzctDomainClearVolumes(dom->def, env);

    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSnapshotDelete(virDomainSnapshotPtr snapshot,
                         unsigned int flags)
{
    virDomainMomentObjPtr snap = NULL;
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    bool job = false;
    int ret = -1;
    int err;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotDeleteEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (!virDomainObjIsActive(dom) &&
        vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    VZCT_CHECK_LOCKED(cleanup, env, "deleting snapshot",
                      vzctl2_env_delete_snapshot(env, snap->def->name));

    if ((snap = vzctSnapObjFromSnapshot(dom->snapshots, snapshot))) {
        g_autofree char *path = NULL;

        path = g_strdup_printf("%s/"VZCT_SNAPHOTS_DIR"/%s.xml",
                               priv->bundlePath, snap->def->name);
        virDomainSnapshotObjListRemove(dom->snapshots, snap);
        if (unlink(path) < 0)
            VIR_WARN("Failed to unlink path=%s, errno=%d", path, errno);
    }

    ret = 0;

 cleanup:
    if (env && !virDomainObjIsActive(dom))
        vzctDomainClearVolumes(dom->def, env);
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}

/*
 * VIR_MIGRATE_NON_SHARED_DISK and VIR_MIGRATE_PARAM_MIGRATE_DISKS are
 * currently not really taken into account although allowed in API. In libvirt
 * migration API the client is responsible to specify which disks are shared
 * and need NOT to be copied and which are NON shared and need to be copied.
 * vzmigrate utility which is used for migration in turn make this desicions by
 * itself. vzmigrate desicions currently are not visible and cannot be sanity
 * checked against client specifications. However if both vzmigrate and client
 * are sane they should have same desicions in this respect.
 */

#define VZCT_MIGRATION_FLAGS \
    VIR_MIGRATE_LIVE | \
    VIR_MIGRATE_PERSIST_DEST | \
    VIR_MIGRATE_UNDEFINE_SOURCE | \
    VIR_MIGRATE_NON_SHARED_DISK | \
    VIR_MIGRATE_CHANGE_PROTECTION | \
    0

#define VZCT_MIGRATION_PARAMETERS \
    VIR_MIGRATE_PARAM_DEST_XML, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_URI, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_DEST_NAME, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_MIGRATE_DISKS, VIR_TYPED_PARAM_STRING | \
                                     VIR_TYPED_PARAM_MULTIPLE, \
    NULL

typedef struct _vzctMigrationCookie vzctMigrationCookie;
typedef vzctMigrationCookie *vzctMigrationCookiePtr;
struct _vzctMigrationCookie {
    char *ctid;

    virHashTablePtr pools;
};

static void
vzctMigrationCookieFree(vzctMigrationCookiePtr cookie)
{
    if (!cookie)
        return;

    virHashFree(cookie->pools);
    cookie->pools = NULL;
    VIR_FREE(cookie->ctid);
    VIR_FREE(cookie);
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(vzctMigrationCookie, vzctMigrationCookieFree);


static void
vzctBakeCookie(vzctMigrationCookiePtr cookie,
               char **cookieout,
               int *cookieoutlen)
{
    g_auto(virBuffer) buf = VIR_BUFFER_INITIALIZER;

    virBufferAddLit(&buf, "<vzct-migration>\n");
    virBufferAdjustIndent(&buf, 2);

    if (cookie->ctid)
        virBufferAsprintf(&buf, "<ctid>%s</ctid>\n", cookie->ctid);

    if (cookie->pools) {
        g_autofree virHashKeyValuePair *pools = NULL;
        size_t i;

        virBufferAddLit(&buf, "<volume-pools>\n");
        virBufferAdjustIndent(&buf, 2);
        pools = virHashGetItems(cookie->pools, NULL);
        for (i = 0; i < virHashSize(cookie->pools); i++)
            virBufferAsprintf(&buf, "<pool name='%s' path='%s'/>\n",
                              (const char *)pools[i].key,
                              (const char *)pools[i].value);
        virBufferAdjustIndent(&buf, -2);
        virBufferAddLit(&buf, "</volume-pools>\n");
    }

    virBufferAdjustIndent(&buf, -2);
    virBufferAddLit(&buf, "</vzct-migration>");

    *cookieout = virBufferContentAndReset(&buf);
    *cookieoutlen = strlen(*cookieout) + 1;
}


static vzctMigrationCookiePtr
vzctEatCookie(const char *cookiein,
              int cookieinlen)
{
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    g_autoptr(xmlDoc) doc = NULL;
    g_autoptr(xmlXPathContext) ctx = NULL;
    g_autofree xmlNodePtr *nodes = NULL;
    size_t i;
    int n;

    VIR_ALLOC(cookie);

    if (!cookiein || cookieinlen == 0)
        return g_steal_pointer(&cookie);

    if (cookiein[cookieinlen - 1] != '\0') {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Invalid migration cookie"));
        return NULL;
    }

    if (!(doc = virXMLParseStringCtxt(cookiein,
                                      _("(_migration_cookie)"), &ctx)))
        return NULL;

    cookie->ctid = virXPathString("string(./ctid[1])", ctx);

    if ((n = virXPathNodeSet("./volume-pools/pool", ctx, &nodes)) < 0)
        return NULL;

    if (n > 0 && !(cookie->pools = virHashCreate(n, virHashValueFree)))
        return NULL;

    for (i = 0; i < n; i++) {
        g_autofree char *name = NULL;
        g_autofree char *path = NULL;

        if (!(name = virXMLPropString(nodes[i], "name"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("missing pool name"));
            return NULL;
        }

        /*
         * Future proof. Don't fail if path is missing. May be in future
         * we will deal with pools that does not have this attribute.
         */
        path = virXMLPropString(nodes[i], "path");
        if (path && virHashAddEntry(cookie->pools, name, path) < 0)
            return NULL;
        path = NULL;
    }

    return g_steal_pointer(&cookie);
}



static char *
vzctDomainMigrateBeginStep(vzctDriverPtr driver,
                           virDomainObjPtr dom,
                           virTypedParameterPtr params,
                           int nparams,
                           unsigned int flags)
{
    int nmigrate_disks = 0;
    const char **migrate_disks = NULL;
    size_t i;

    if ((flags & VIR_MIGRATE_LIVE) &&
        !virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("live migration only makes sense for active domains"));
        return NULL;
    }

    if (!(flags & VIR_MIGRATE_UNDEFINE_SOURCE)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("you need to set flag to undefine source because"
                         " not undefining is not supported"));
        return NULL;
    }

    if (!(flags & VIR_MIGRATE_PERSIST_DEST)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("you need to set flag to persist on destination because"
                         " not persisting is not supported"));
        return NULL;
    }

    if (virTypedParamsGet(params, nparams, VIR_MIGRATE_PARAM_DEST_XML)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("changing xml on migration is not supported"));
        return NULL;
    }


    /* params can be NULL and virTypedParamsGetStringList error out on NULL */
    if (params) {
        nmigrate_disks = virTypedParamsGetStringList(params, nparams,
                                                     VIR_MIGRATE_PARAM_MIGRATE_DISKS,
                                                     &migrate_disks);
        if (nmigrate_disks < 0)
            return NULL;
    }

    for (i = 0; i < nmigrate_disks; i++) {
        size_t j;

        for (j = 0; j < dom->def->ndisks; j++) {
            virDomainDiskDefPtr disk = dom->def->disks[j];

            if (STREQ(migrate_disks[i], disk->dst)) {
                if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
                    virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                                   _("cannot migrate non disk devices"));
                    return NULL;
                }

                break;
            }
        }

        if (j == dom->def->ndisks) {
            virReportError(VIR_ERR_INVALID_ARG,
                           _("cannot find disk '%s' specified for copy"),
                           migrate_disks[i]);
            return NULL;
        }
    }

    if (virDomainObjIsActive(dom)) {
        g_autofree char *migfile = vzctMigFile(dom);

        if (virDomainObjSavePath(dom, driver->xmlopt, migfile) < 0)
            return NULL;
    }

    return virDomainDefFormat(dom->def, driver->xmlopt,
                              VIR_DOMAIN_DEF_FORMAT_MIGRATABLE |
                              VIR_DOMAIN_DEF_FORMAT_SECURE |
                              VIR_DOMAIN_DEF_FORMAT_INACTIVE);
}


static char *
vzctDomainMigrateBegin3Params(virDomainPtr domain,
                              virTypedParameterPtr params,
                              int nparams,
                              char **cookieout G_GNUC_UNUSED,
                              int *cookieoutlen G_GNUC_UNUSED,
                              unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom;
    char* xml = NULL;
    bool job = false;

    virCheckFlags(VZCT_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return NULL;

    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("change protection flag must be set"));

        return NULL;
    };

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MIGRATION) < 0)
        goto cleanup;
    job = true;
    priv = dom->privateData;
    priv->mig_conn_id = domain->conn->id;

    if (!(xml = vzctDomainMigrateBeginStep(driver, dom, params, nparams, flags)))
        goto cleanup;

    vzctDomainAllowSideJob(dom);

 cleanup:
    if (job && !xml)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return xml;
}


static char*
vzctMigrationCreateURI(void)
{
    g_autofree char *hostname = NULL;
    char *uri;

    if (!(hostname = virGetHostname()))
        return NULL;

    if (STRPREFIX(hostname, "localhost")) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("hostname on destination resolved to localhost,"
                         " but migration requires an FQDN"));
        return NULL;
    }

    uri = g_strdup_printf("vzctmigr://%s", hostname);
    return uri;
}


static int
vzctDomainDstCookieAddPools(virConnectPtr conn,
                            virDomainDefPtr def,
                            vzctMigrationCookiePtr cookie)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        g_autoptr(virStoragePool) pool = NULL;
        g_autoptr(virStoragePoolDef) pooldef = NULL;
        g_autofree char *poolxml = NULL;
        g_autofree char *path = NULL;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        if (!(pool = virStoragePoolLookupByName(conn, disk->src->srcpool->pool)))
            return -1;

        if (!cookie->pools &&
            !(cookie->pools = virHashCreate(10, virHashValueFree)))
            return -1;

        if (virHashLookup(cookie->pools, disk->src->srcpool->pool))
            continue;

        if (!(poolxml = virStoragePoolGetXMLDesc(pool, 0)))
            return -1;

        if (!(pooldef = virStoragePoolDefParseString(poolxml)))
            return -1;

        switch (pooldef->type) {
        case VIR_STORAGE_POOL_DIR:
        case VIR_STORAGE_POOL_FS:
        case VIR_STORAGE_POOL_NETFS:
        case VIR_STORAGE_POOL_VSTORAGE:
            path = g_strdup(pooldef->target.path);
            if (virHashAddEntry(cookie->pools, pooldef->name, path) < 0)
                return -1;
            path = NULL;
            break;

        default:
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                           _("pool type '%s' is not supported"),
                           virStoragePoolTypeToString(pooldef->type));
            return -1;
        }
    }

    return 0;
}


static int
vzctDomainMigratePrepare3Params(virConnectPtr conn,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein G_GNUC_UNUSED,
                                int cookieinlen G_GNUC_UNUSED,
                                char **cookieout,
                                int *cookieoutlen,
                                char **uri_out,
                                unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    unsigned parseFlags = VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                          VIR_DOMAIN_DEF_PARSE_INACTIVE;
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    const char *xml = NULL;
    const char *uri = NULL;

    virCheckFlags(VZCT_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_XML,
                                &xml) < 0)
        return -1;

    if (!xml) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("missing destination xml parameter"));
        return -1;
    }

    if (!(flags & VIR_MIGRATE_PERSIST_DEST)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("you need to set flag to persist on destination because"
                         " not persisting is not supported"));
        return -1;
    }

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_URI, &uri) < 0)
        return -1;

    if (!uri && !(*uri_out = vzctMigrationCreateURI()))
        return -1;

    if (!(def = virDomainDefParseString(xml, driver->xmlopt,
                                        NULL, parseFlags)))
        return -1;

    VIR_ALLOC(cookie);
    if (vzctDomainDstCookieAddPools(conn, def, cookie) < 0)
        return -1;

    vzctBakeCookie(cookie, cookieout, cookieoutlen);
    return 0;
}


#define VZCT_VZMIGRATE_BIN "/usr/sbin/vzmigrate"

static int
vzctMigrateVz(virDomainObjPtr dom,
              const char *uristr,
              const char *dname,
              unsigned int flags)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    g_autoptr(virURI) uri = NULL;
    g_autoptr(virCommand) cmd = NULL;
    pid_t pid;
    int status;
    int rc;

    if (!(uri = virURIParse(uristr)))
        return -1;

    if (!virFileIsExecutable(VZCT_VZMIGRATE_BIN)) {
        virReportSystemError(errno,
                             _("Cannot find %s - "
                               "Possibly appopriate package isn't installed"),
                             VZCT_VZMIGRATE_BIN);
        return -1;
    }

    cmd = virCommandNew(VZCT_VZMIGRATE_BIN);
    virCommandAddArg(cmd, uri->server);
    virCommandAddArg(cmd, dom->def->extraId);

    if (flags & VIR_MIGRATE_LIVE)
        virCommandAddArgList(cmd, "--online", NULL);
    if (dname)
        virCommandAddArgList(cmd, "--new-name", dname, NULL);

    if (virCommandRunAsync(cmd, &priv->migrationPid) < 0)
        return -1;

    pid = priv->migrationPid;
    vzctDomainAllowSideJob(dom);
    virObjectUnlock(dom);
    rc = virProcessWait(pid, &status, true);
    virObjectLock(dom);
    vzctDomainDisableSideJob(dom);
    priv->migrationPid = 0;

    if (rc < 0)
        return -1;

    if (!(WIFEXITED(status) && WEXITSTATUS(status) == 0)) {
        if (priv->abortJob) {
            virReportError(VIR_ERR_OPERATION_ABORTED, "%s",
                           _("migration cancelled by client"));
        } else {
            g_autofree char *st = virProcessTranslateStatus(status);

            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Migration process unexpected %s"), st);
        }
        return -1;
    }

    return rc;
}


static int
vzctDomainSrcCheckPools(virDomainDefPtr def,
                        vzctMigrationCookiePtr cookie)
{
    g_autoptr(virConnect) conn = NULL;
    size_t i;

    if (!(conn = virGetConnectStorage()))
        return -1;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        g_autoptr(virStoragePool) pool = NULL;
        g_autoptr(virStoragePoolDef) pooldef = NULL;
        g_autofree char *poolxml = NULL;
        const char *path;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        if (!cookie->pools) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("missing pools info from destination"));
            return -1;
        }

        if (!(path = virHashLookup(cookie->pools, disk->src->srcpool->pool))) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("missing info for pool '%s' from destination"),
                           disk->src->srcpool->pool);
            return -1;
        }

        if (!(pool = virStoragePoolLookupByName(conn, disk->src->srcpool->pool)))
            return -1;

        if (!(poolxml = virStoragePoolGetXMLDesc(pool, 0)))
            return -1;

        if (!(pooldef = virStoragePoolDefParseString(poolxml)))
            return -1;

        switch (pooldef->type) {
        case VIR_STORAGE_POOL_DIR:
        case VIR_STORAGE_POOL_FS:
        case VIR_STORAGE_POOL_NETFS:
        case VIR_STORAGE_POOL_VSTORAGE:
            if (STRNEQ(path, pooldef->target.path)) {
                virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                               _("pool '%s' have different path on src and dst"),
                               pooldef->name);
                return -1;
            }
            break;

        default:
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                           _("pool type '%s' is not supported"),
                           virStoragePoolTypeToString(pooldef->type));
            return -1;
        }
    }

    return 0;
}


static int
vzctDomainMigratePerformStep(vzctDriverPtr driver,
                             virDomainObjPtr dom,
                             virTypedParameterPtr params,
                             int nparams,
                             const char *cookiein,
                             int cookieinlen,
                             char **cookieout,
                             int *cookieoutlen,
                             unsigned int flags)
{
    virObjectEventPtr event;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    g_autoptr(vzctMigrationCookie) incookie = NULL;
    const char *uri = NULL;
    const char *dname = NULL;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_URI, &uri) < 0 ||
        virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_NAME, &dname) < 0)
        return -1;

    if (!uri) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("missing migration uri"));
        return -1;
    }

    if (!(incookie = vzctEatCookie(cookiein, cookieinlen)))
        return -1;

    if (vzctDomainSrcCheckPools(dom->def, incookie) < 0)
        return -1;

    if (!virDomainObjIsActive(dom)) {
        g_autoptr(vzctl_env_handle) env = NULL;
        int err;

        VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
        if (vzctDomainResolveVolumes(dom->def, env) < 0)
            return -1;
    }

    if (vzctMigrateVz(dom, uri, dname, flags) < 0) {
        g_autofree char *migfile = vzctMigFile(dom);

        if (unlink(migfile) < 0 && errno != ENOENT)
            VIR_WARN("Cannot remove migrating status file: %s, error: %d",
                     migfile, errno);

        if (!virDomainObjIsActive(dom)) {
            g_autoptr(vzctl_env_handle) env = NULL;
            int err;

            VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
            vzctDomainClearVolumes(dom->def, env);
        }

        return -1;
    }

    if (virDomainObjIsActive(dom)) {
        vzctDomainStopFlags(driver, dom, true);
        event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_DESTROYED);
        virObjectEventStateQueue(driver->eventsManager, event);
    }

    virDomainObjListRemove(driver->domains, dom);
    VIR_ALLOC(cookie);
    cookie->ctid = dom->def->extraId;
    vzctBakeCookie(cookie, cookieout, cookieoutlen);
    event = virDomainEventLifecycleNewFromObj(dom,
                                 VIR_DOMAIN_EVENT_UNDEFINED,
                                 VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);
    virObjectEventStateQueue(driver->eventsManager, event);

    return 0;
}


static int
vzctDomainMigrateConfirmStep(virDomainObjPtr dom,
                             int cancelled);

/*
 * On success steps order is next with begin step on source side
 * and each next step on the other side:
 * - begin/prepare/perform/finish/confirm
 *
 * On any error if the last step on a side was successful then last
 * step of success order on that side is executed with cancelled
 * flag set to 1. And if both finish and confirm have to be
 * executed then finish is executed first.
 *
 * Cookie is passed from step to step but on error the cleanup step
 * SHOULD not use cookie as its content is unspecified.
 *
 */
static int
vzctDomainMigratePerformP2P(vzctDriverPtr driver,
                            virDomainObjPtr dom,
                            const char *dconnuri,
                            virTypedParameterPtr params,
                            int nparams,
                            unsigned int flags)
{
    virErrorPtr orig_err = NULL;
    g_autofree char *uri = NULL;
    g_autofree char *cookiein = NULL;
    g_autofree char *cookieout = NULL;
    g_autofree char *xml = NULL;
    g_autoptr(virConnect) dconn = NULL;
    g_autoptr(virDomain) ddomain = NULL;
    int cookieinlen = 0;
    int cookieoutlen = 0;
    bool confirm = true;
    virTypedParameterPtr tmp;
    int ret = -1;
    int rc;

    virCheckFlags(VZCT_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (virTypedParamsCopy(&tmp, params, nparams) < 0)
        return -1;
    params = tmp;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MIGRATION) < 0)
        goto cleanup;

    if (!(dconn = virConnectOpen(dconnuri)))
        goto cleanup;

    if (!(xml = vzctDomainMigrateBeginStep(driver, dom, params, nparams, flags)))
        goto cleanup;

    virTypedParamsReplaceString(&params, &nparams,
                                VIR_MIGRATE_PARAM_DEST_XML, xml);

    cookiein = g_steal_pointer(&cookieout);
    cookieinlen = cookieoutlen;
    cookieoutlen = 0;

    vzctDomainAllowSideJob(dom);
    virObjectUnlock(dom);
    rc = dconn->driver->domainMigratePrepare3Params
            (dconn, params, nparams, cookiein, cookieinlen,
             &cookieout, &cookieoutlen, &uri, flags);
    virObjectLock(dom);
    vzctDomainDisableSideJob(dom);
    if (rc < 0)
        goto confirm;

    if (uri)
        virTypedParamsReplaceString(&params, &nparams,
                                    VIR_MIGRATE_PARAM_URI, uri);

    VIR_FREE(cookiein);
    cookiein = g_steal_pointer(&cookieout);
    cookieinlen = cookieoutlen;
    cookieoutlen = 0;

    if (vzctDomainMigratePerformStep(driver, dom, params, nparams,
                                     cookiein, cookieinlen,
                                     &cookieout, &cookieoutlen,
                                     flags) < 0) {
        virErrorPreserveLast(&orig_err);
        confirm = false;
        goto finish;
    }

    ret = 0;

 finish:
    VIR_FREE(cookiein);
    cookiein = g_steal_pointer(&cookieout);
    cookieinlen = cookieoutlen;
    cookieoutlen = 0;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_NAME, NULL) == 0)
        virTypedParamsReplaceString(&params, &nparams,
                                    VIR_MIGRATE_PARAM_DEST_NAME,
                                    dom->def->name);
    vzctDomainAllowSideJob(dom);
    virObjectUnlock(dom);
    ddomain = dconn->driver->domainMigrateFinish3Params(dconn, params, nparams,
                                                        cookiein, cookieinlen,
                                                        &cookieout, &cookieoutlen,
                                                        flags,
                                                        ret == 0 ? 0 : 1);
    virObjectLock(dom);
    vzctDomainDisableSideJob(dom);
    if (ret == 0 && !ddomain)
        ret = -1;

 confirm:
    if (confirm)
        vzctDomainMigrateConfirmStep(dom, ret == 0 ? 0 : 1);

 cleanup:
    vzctDomainEndJob(dom);
    virTypedParamsFree(params, nparams);
    virErrorRestore(&orig_err);

    return ret;
}


static int
vzctDomainMigratePerformManaged(vzctDriverPtr driver,
                                virDomainObjPtr dom,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein,
                                int cookieinlen,
                                char **cookieout,
                                int *cookieoutlen,
                                unsigned int flags)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    int ret = -1;

    if (priv->job != VZCT_DOMAIN_JOB_MIGRATION) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("migration perform step without active migration"));
        return -1;
    }

    /*
     * Check even flags and params after job presence is established
     * so in case of bad parameters we can cleanup job.
     */
    if (flags & ~(VZCT_MIGRATION_FLAGS)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("unsupported flags"));
        goto cleanup;
    }

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        goto cleanup;

    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("change protection flag must be set"));
        goto cleanup;
    };

    vzctDomainDisableSideJob(dom);
    if (vzctDomainMigratePerformStep(driver, dom, params, nparams,
                                     cookiein, cookieinlen,
                                     cookieout, cookieoutlen,
                                     flags) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    /*
     * End job for success path too as we remove domain on successful migration
     * on perform step and confirm step has no chances to find domain in list
     * and end migration job.
     */
    vzctDomainEndJob(dom);

    return ret;
}


static int
vzctDomainMigratePerform3Params(virDomainPtr domain,
                                const char *dconnuri G_GNUC_UNUSED,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein,
                                int cookieinlen,
                                char **cookieout,
                                int *cookieoutlen,
                                unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom;
    int ret;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (flags & VIR_MIGRATE_PEER2PEER)
        ret = vzctDomainMigratePerformP2P(driver, dom, dconnuri,
                                          params, nparams,
                                          flags & (~VIR_MIGRATE_PEER2PEER));
    else
        ret = vzctDomainMigratePerformManaged(driver, dom,
                                              params, nparams,
                                              cookiein, cookieinlen,
                                              cookieout, cookieoutlen,
                                              flags);

    virDomainObjEndAPI(&dom);
    return ret;
}


static virDomainPtr
vzctDomainMigrateFinish3Params(virConnectPtr dconn,
                               virTypedParameterPtr params,
                               int nparams,
                               const char *cookiein,
                               int cookieinlen,
                               char **cookieout G_GNUC_UNUSED,
                               int *cookieoutlen G_GNUC_UNUSED,
                               unsigned int flags,
                               int cancelled)
{
    vzctDriverPtr driver = dconn->privateData;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    virDomainObjPtr dom = NULL;
    virDomainPtr domain = NULL;

    virCheckFlags(VZCT_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return NULL;

    if (cancelled) {
        /* need to report some error */
        virReportError(VIR_ERR_MIGRATE_FINISH_OK, NULL);
        return NULL;
    }

    if (!(cookie = vzctEatCookie(cookiein, cookieinlen)))
        return NULL;

    if (!cookie->ctid) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("missing ctid"));
        return NULL;
    }

    if (!(dom = vzctAddDomain(driver, cookie->ctid, NULL)))
        return NULL;

    if (dom->state.state == VIR_DOMAIN_NOSTATE) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("domain is out of sync on destination"));
        goto cleanup;
    }

    if (!virDomainObjIsActive(dom)) {
        g_autoptr(vzctl_env_handle) env = NULL;
        int err;

        VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));
        vzctDomainClearVolumes(dom->def, env);
    }

    domain = virGetDomain(dconn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    virDomainObjEndAPI(&dom);
    return domain;
}


static int
vzctDomainMigrateConfirmStep(virDomainObjPtr dom,
                             int cancelled)
{
    if (cancelled) {
        vzctl_env_status_t status;

        VZCT_CHECK_RET(0, vzctl2_get_env_status(dom->def->extraId, &status,
                                                ENV_STATUS_RUNNING));

        /*
         * Check status before deleting the file. The problem is we can
         * get cancelled flag set if perform was successful and finish
         * was failed and private directory is shared one. In this case
         * we don't want to mess with destination which will use this
         * file.
         */
        if (status.mask & ENV_STATUS_RUNNING) {
            g_autofree char *migfile = vzctMigFile(dom);

            if (unlink(migfile) < 0 && errno != ENOENT)
                VIR_WARN("Cannot remove migrating status file: %s, error: %d",
                         migfile, errno);
        }
    }
    return 0;
}


static int
vzctDomainMigrateConfirm3Params(virDomainPtr domain,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein G_GNUC_UNUSED,
                                int cookieinlen G_GNUC_UNUSED,
                                unsigned int flags,
                                int cancelled)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom;
    bool job = false;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return 0;

    priv = dom->privateData;
    if (priv->job != VZCT_DOMAIN_JOB_MIGRATION) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("migration confirm step without active migration"));
        goto cleanup;
    }
    job = true;

    /*
     * Check even flags and params after job presence is established
     * so in case of bad parameters we can cleanup job.
     */
    if (flags & ~(VZCT_MIGRATION_FLAGS)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("unsupported flags"));
        goto cleanup;
    }

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        goto cleanup;

    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("change protection flag must be set"));
        goto cleanup;
    };

    vzctDomainDisableSideJob(dom);
    vzctDomainMigrateConfirmStep(dom, cancelled);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return 0;
}


static int
vzctConnectDomainEventRegisterAny(virConnectPtr conn,
                                  virDomainPtr dom,
                                  int eventID,
                                  virConnectDomainEventGenericCallback callback,
                                  void *opaque,
                                  virFreeCallback freecb)
{
    vzctDriverPtr driver = conn->privateData;
    int ret = -1;

    if (virConnectDomainEventRegisterAnyEnsureACL(conn) < 0)
        return -1;

    if (virDomainEventStateRegisterID(conn,
                                      driver->eventsManager,
                                      dom, eventID,
                                      callback, opaque, freecb, &ret) < 0)
        return -1;

    return ret;
}


static int
vzctConnectDomainEventDeregisterAny(virConnectPtr conn,
                                    int callbackID)
{
    vzctDriverPtr driver = conn->privateData;

    if (virConnectDomainEventDeregisterAnyEnsureACL(conn) < 0)
        return -1;

    if (virObjectEventStateDeregisterID(conn,
                                        driver->eventsManager,
                                        callbackID, true) < 0)
        return -1;

    return 0;
}


static int
vzctDomainGetJobStatsInternal(virDomainObjPtr dom,
                              vzctDomainJobStatsPtr *ret)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    g_autofree vzctDomainJobStats *stats = NULL;
    unsigned long long now;

    if (priv->job != VZCT_DOMAIN_JOB_MIGRATION) {
        *ret = NULL;
        return 0;
    }

    if (virTimeMillisNow(&now) < 0)
        return -1;

    VIR_ALLOC(stats);
    stats->timeElapsed = now - priv->jobStarted;

    *ret = g_steal_pointer(&stats);
    return 0;
}


static int
vzctDomainGetJobInfo(virDomainPtr domain,
                     virDomainJobInfoPtr info)
{
    g_autofree vzctDomainJobStats *stats = NULL;
    virDomainObjPtr dom;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetJobInfoEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainGetJobStatsInternal(dom, &stats) < 0)
        goto cleanup;

    memset(info, 0, sizeof(*info));
    if (stats) {
        info->type = VIR_DOMAIN_JOB_UNBOUNDED;
        info->timeElapsed = stats->timeElapsed;
    } else {
        info->type = VIR_DOMAIN_JOB_NONE;
    }
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctConvertStatsToParams(vzctDomainJobStatsPtr stats,
                         virTypedParameterPtr *params_out,
                         int *nparams_out)
{
    virTypedParameterPtr params = NULL;
    int maxparams = 0;
    int nparams = 0;
    int ret = -1;

    if (!stats) {
        *params_out = NULL;
        *nparams_out = 0;
        return 0;
    }

    if (virTypedParamsAddInt(&params, &nparams, &maxparams,
                             VIR_DOMAIN_JOB_OPERATION,
                             VIR_DOMAIN_JOB_OPERATION_MIGRATION_OUT) < 0)
        goto cleanup;

    if (virTypedParamsAddULLong(&params, &nparams, &maxparams,
                                VIR_DOMAIN_JOB_TIME_ELAPSED,
                                stats->timeElapsed) < 0)
        goto cleanup;

    *params_out = g_steal_pointer(&params);
    *nparams_out = nparams;
    nparams = 0;
    ret = 0;

 cleanup:
    virTypedParamsFree(params, nparams);
    return ret;
}


static int
vzctDomainGetJobStats(virDomainPtr domain,
                      int *type,
                      virTypedParameterPtr *params,
                      int *nparams,
                      unsigned int flags)
{
    g_autofree vzctDomainJobStats *stats = NULL;
    virDomainObjPtr dom;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetJobStatsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainGetJobStatsInternal(dom, &stats) < 0)
        goto cleanup;

    if (vzctConvertStatsToParams(stats, params, nparams) < 0)
        goto cleanup;

    *type = stats ? VIR_DOMAIN_JOB_UNBOUNDED : VIR_DOMAIN_JOB_NONE;
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainAbortJob(virDomainPtr domain)
{
    virDomainObjPtr dom;
    vzctDomainObjPrivatePtr priv;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainAbortJobEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!priv->migrationPid) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       _("domain %s does not have active migration"),
                       dom->def->name);
        goto cleanup;
    }

    virProcessKill(priv->migrationPid, SIGTERM);
    priv->abortJob = true;

    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainGetAutostart(virDomainPtr domain,
                       int *autostart)
{
    virDomainObjPtr dom = NULL;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetAutostartEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    *autostart = dom->autostart;
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSetAutostart(virDomainPtr domain,
                       int autostart)
{
    g_autofree char *autostartFile = NULL;
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    bool job = false;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetAutostartEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    autostart = !!autostart;
    if (dom->autostart == autostart) {
        ret = 0;
        goto cleanup;
    }

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    priv = dom->privateData;
    autostartFile = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/autostart", priv->bundlePath);

    if (autostart) {
        if (virFileTouch(autostartFile, 0600)) {
            virReportSystemError(errno,
                                 _("Failed to create autostart symlink '%s'"),
                                 autostartFile);
            goto cleanup;
        }
    } else {
        if (unlink(autostartFile) < 0 && errno != ENOENT) {
            virReportSystemError(errno,
                                 _("Failed to delete autostart symlink '%s'"),
                                 autostartFile);
            goto cleanup;
        }
    }

    dom->autostart = autostart;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCommandXExec(virDomainPtr domain,
                       const char *cmd,
                       const char **args,
                       unsigned int nargs,
                       const char **envs,
                       unsigned int nenvs,
                       virTypedParameterPtr params,
                       int nparams,
                       virStreamPtr in,
                       virStreamPtr out,
                       virStreamPtr err,
                       unsigned int flags)
{
    virDomainObjPtr dom = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    VIR_AUTOSTRINGLIST argv = NULL;
    VIR_AUTOSTRINGLIST envp = NULL;
    vzctDomainObjPrivatePtr priv;
    vzctDomainExecProcessPtr exec_proc;
    struct vzctl_exec_handle *exec_handle;
    struct winsize ws;
    bool job = false;
    bool pty = !!(flags & VIR_DOMAIN_COMMAND_X_EXEC_TERMINAL);
    int fds[3];
    int pipes_in[2] = { -1, -1 };
    int pipes_out[2] = { -1, -1 };
    int pipes_err[2] = { -1, -1 };
    int vz_err;
    size_t i;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_COMMAND_X_EXEC_SHELL |
                  VIR_DOMAIN_COMMAND_X_EXEC_TERMINAL, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                               VIR_TYPED_PARAM_UINT,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                               VIR_TYPED_PARAM_UINT,
                               NULL) < 0)
        return -1;

    if (pty) {
        int rows;
        int columns;

        if (err) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("stderr is not available in terminal mode"));
            return -1;
        }

        if (virTypedParamsGetUInt(params, nparams,
                                  VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                                  (unsigned int *)&rows) < 0)
            return -1;

        if (virTypedParamsGetUInt(params, nparams,
                                  VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                                  (unsigned int *)&columns) < 0)
            return -1;

        if ((rows == -1 && columns >= 0) ||
            (columns == -1 && rows >= 0)) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("rows and columns should be either"
                             " both set or both unset"));
            return -1;
        }

        memset(&ws, 0, sizeof(ws));
        ws.ws_row = rows;
        ws.ws_col = columns;

        if (flags & VIR_DOMAIN_COMMAND_X_EXEC_SHELL) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("'shell' flag does not make sense with"
                             " 'terminal' flag"));
            return -1;
        }
    } else {
        if (virTypedParamsGet(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS) ||
            virTypedParamsGet(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS)) {

            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("window size parameters does not make sense "
                             "without terminal flag"));
            return -1;
        }
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXExecEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                   VZCTL_CONF_SKIP_PARSE, &vz_err));

    VIR_ALLOC_N(argv, nargs + 2);
    argv[0] = g_strdup(cmd);
    for (i = 0; i < nargs; i++)
        argv[i + 1] = g_strdup(args[i]);

    VIR_ALLOC_N(envp, nenvs + 1);
    for (i = 0; i < nenvs; i++)
        envp[1] = g_strdup(envs[i]);

    /*
     * Use sockets instead of pipes due to the way streams are implemented
     * in rpc. On receiving finish/abort on stdout stream that does not
     * receive EOF yet rpc waits for writable event which will never
     * come if stdout is read end of pipe. Thus ^C on next command hangs:
     *
     * virsh -k0 -c vzct:///system x-exec 100 sleep 600
     */
    if (socketpair(PF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, pipes_in) < 0 ||
        socketpair(PF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, pipes_out) < 0) {
        virReportSystemError(errno, "%s", _("socketpair() failed"));
        goto cleanup;
    }
    fds[0] = pipes_in[0];
    fds[1] = pipes_out[1];
    if (virFDStreamOpen(in, pipes_in[1]) < 0)
        goto cleanup;
    pipes_in[1] = -2;
    if (virFDStreamOpen(out, pipes_out[0]) < 0)
        goto cleanup;
    pipes_out[0] = -2;

    if (pty) {
        VZCT_CHECK(cleanup,
                   vzctl2_env_exec_pty(env, argv, envp, fds,
                                       NULL, &ws, &exec_handle));
    } else {
        exec_mode_e mode;

        if (flags & VIR_DOMAIN_COMMAND_X_EXEC_SHELL)
            mode = MODE_BASH;
        else
            mode = MODE_EXEC;

        if (socketpair(PF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, pipes_err) < 0) {
            virReportSystemError(errno, "%s", _("socketpair() failed"));
            goto cleanup;
        }
        fds[2] = pipes_err[1];
        if (virFDStreamOpen(err, pipes_err[0]) < 0)
            goto cleanup;
        pipes_err[0] = -2;

        VZCT_CHECK(cleanup,
                   vzctl2_env_execve(env, argv, envp, fds, mode, &exec_handle));
    }

    priv = dom->privateData;
    VIR_ALLOC(exec_proc);
    exec_proc->handle = exec_handle;
    exec_proc->conn_id = domain->conn->id;
    VIR_RESIZE_N(priv->exec_procs,
                 priv->exec_nprocs_max, priv->exec_nprocs, 1);
    priv->exec_procs[priv->exec_nprocs++] = exec_proc;

    ret = exec_handle->pid;

 cleanup:
    if (ret < 0) {
        virErrorPtr saveerr = virSaveLastError();

        if (pipes_in[1] == -2)
            virStreamAbort(in);
        if (pipes_out[0] == -2)
            virStreamAbort(out);
        if (pipes_err[0] == -2)
            virStreamAbort(err);

        virSetError(saveerr);
        virFreeError(saveerr);
    }

    VIR_FORCE_CLOSE(pipes_in[0]);
    VIR_FORCE_CLOSE(pipes_in[1]);
    VIR_FORCE_CLOSE(pipes_out[0]);
    VIR_FORCE_CLOSE(pipes_out[1]);
    VIR_FORCE_CLOSE(pipes_err[0]);
    VIR_FORCE_CLOSE(pipes_err[1]);

    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSearchExecHandle(virDomainObjPtr dom,
                           int pid_value)
{
    vzctDomainObjPrivatePtr priv;
    size_t i;

    priv = dom->privateData;
    for (i = 0; i < priv->exec_nprocs; i++) {
        if (priv->exec_procs[i]->handle->pid == pid_value)
            return i;
    }

    virReportError(VIR_ERR_INVALID_ARG, _("invalid pid %d"), pid_value);
    return -1;
}


static int
vzctDomainCommandXGetStatus(virDomainPtr domain,
                            int pid_value,
                            virDomainCommandXStatusPtr status,
                            unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    bool job = false;
    int wstatus;
    int ret = -1;
    int idx;
    int rc;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXGetStatusEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if ((idx = vzctDomainSearchExecHandle(dom, pid_value)) < 0)
        goto cleanup;

    priv = dom->privateData;
    rc = vzctl2_env_waitpid(priv->exec_procs[idx]->handle, 1, &wstatus);
    if (rc < 0) {
        const char *err = vzctl2_get_last_error();
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("vzctl2_env_waitpid failed with error %s"),
                       NULLSTR(err));
        goto cleanup;
    }

    status->exited = 0;
    status->code = -1;
    status->signal = -1;
    if (rc == 0) {
        ret = 0;
        goto cleanup;
    }
    status->exited = 1;
    if (WIFEXITED(wstatus)) {
        status->code = WEXITSTATUS(wstatus);
    } else {
        status->signal = WTERMSIG(wstatus);
    }

    vzctl2_release_exec_handle(priv->exec_procs[idx]->handle);
    VIR_DELETE_ELEMENT_INPLACE(priv->exec_procs, idx, priv->exec_nprocs);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCommandXTerminate(virDomainPtr domain,
                            int pid_value,
                            unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    bool job = false;
    int idx;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXTerminateEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if ((idx = vzctDomainSearchExecHandle(dom, pid_value)) < 0)
        goto cleanup;

    priv = dom->privateData;
    vzctl2_env_exec_terminate(priv->exec_procs[idx]->handle);
    priv->exec_procs[idx]->terminated = true;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCommandXSetParams(virDomainPtr domain,
                            int pid_value,
                            virTypedParameterPtr params,
                            int nparams,
                            unsigned int flags)
{
    vzctDomainObjPrivatePtr priv;
    virDomainObjPtr dom = NULL;
    struct winsize ws;
    int rows = -1;
    int columns = -1;
    bool job = false;
    int idx;
    int ret = -1;

    virCheckFlags(0, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                               VIR_TYPED_PARAM_UINT,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                               VIR_TYPED_PARAM_UINT,
                               NULL) < 0)
        return -1;

    if (virTypedParamsGetUInt(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                              (unsigned int *)&rows) < 0)
        return -1;

    if (virTypedParamsGetUInt(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                              (unsigned int *)&columns) < 0)
        return -1;

    if ((rows == -1 && columns >= 0) ||
        (columns == -1 && rows >= 0)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("rows and columns should be either both set or both unset"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXSetParamsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if ((idx = vzctDomainSearchExecHandle(dom, pid_value)) < 0)
        goto cleanup;
    priv = dom->privateData;

    memset(&ws, 0, sizeof(ws));
    ws.ws_col = columns;
    ws.ws_row = rows;

    VZCT_CHECK(cleanup,
               vzctl2_env_exec_set_winsize(priv->exec_procs[idx]->handle, &ws));
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSetMemoryFlags(virDomainPtr domain,
                         unsigned long memory,
                         unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    virObjectEventPtr event;
    bool active = false;
    bool job = false;
    int err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetMemoryFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        if (flags & (!(flags & VIR_DOMAIN_AFFECT_LIVE) ||
                     !(flags & VIR_DOMAIN_AFFECT_CONFIG))) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("Cannot modify independently persistent "
                             "and active configs"));
            goto cleanup;
        }

        active = true;
    } else {
        if (flags & VIR_DOMAIN_AFFECT_LIVE) {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("Cannot modify active config of inactive domain"));
            goto cleanup;
        }
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    /* Ceiling new memory size from Kb to Mb */
    VZCT_CHECK(cleanup, vzctl2_env_set_ramsize(param,
                                               (memory + 1023) / 1024));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (active) {
        dom->newDef->mem.cur_balloon = memory;
        virDomainDefSetMemoryTotal(dom->newDef, dom->newDef->mem.cur_balloon);
    }

    dom->def->mem.cur_balloon = memory;
    virDomainDefSetMemoryTotal(dom->def, dom->def->mem.cur_balloon);

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (active) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetVcpusFlags(virDomainPtr domain,
                        unsigned int nvcpus,
                        unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    virObjectEventPtr event;
    bool active = false;
    bool job = false;
    int err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetVcpusFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        if (flags & (!(flags & VIR_DOMAIN_AFFECT_LIVE) ||
                     !(flags & VIR_DOMAIN_AFFECT_CONFIG))) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("Cannot modify independently persistent "
                             "and active configs"));
            goto cleanup;
        }

        active = true;
    } else {
        if (flags & VIR_DOMAIN_AFFECT_LIVE) {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("Cannot modify active config of inactive domain"));
            goto cleanup;
        }
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    VZCT_CHECK(cleanup, vzctl2_env_set_cpu_count(param, nvcpus));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (active)
        dom->newDef->maxvcpus = nvcpus;
    dom->def->maxvcpus = nvcpus;

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (active) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetVcpusFlags(virDomainPtr domain,
                        unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get simultaneously persistent "
                         "and active configs"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetVcpusFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get active config of inactive domain"));
        goto cleanup;
    }

    ret = virDomainDefGetVcpusMax(dom->def);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetMemoryParameters(virDomainPtr domain,
                              virTypedParameterPtr params,
                              int nparams,
                              unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    struct vzctl_mem_guarantee mem_guarantee;
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    virObjectEventPtr event;
    unsigned long long min_guarantee;
    bool set_min_guarantee = false;
    bool set_min_guarantee_auto = false;
    bool active = false;
    bool job = false;
    int min_guarantee_auto;
    int rc, err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_MEMORY_MIN_GUARANTEE,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_MEMORY_MIN_GUARANTEE_VZ_AUTO,
                               VIR_TYPED_PARAM_BOOLEAN,
                               NULL) < 0)
        return -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetMemoryParametersEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        if (flags & (!(flags & VIR_DOMAIN_AFFECT_LIVE) ||
                     !(flags & VIR_DOMAIN_AFFECT_CONFIG))) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("Cannot modify independently persistent "
                             "and active configs"));
            goto cleanup;
        }

        active = true;
    } else {
        if (flags & VIR_DOMAIN_AFFECT_LIVE) {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("Cannot modify active config of inactive domain"));
            goto cleanup;
        }
    }

    if ((rc = virTypedParamsGetULLong(params, nparams,
                                      VIR_DOMAIN_MEMORY_MIN_GUARANTEE,
                                      &min_guarantee)) < 0)
        goto cleanup;
    if (rc == 1)
        set_min_guarantee = true;

    if ((rc = virTypedParamsGetBoolean(params, nparams,
                                       VIR_DOMAIN_MEMORY_MIN_GUARANTEE_VZ_AUTO,
                                       &min_guarantee_auto)) < 0)
        goto cleanup;
    if (rc == 1)
        set_min_guarantee_auto = true;

    if (set_min_guarantee && set_min_guarantee_auto) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("min_guarantee and min_guarantee_vz_auto are "
                         "mutually exclusive"));
        goto cleanup;
    }

    if (set_min_guarantee_auto && !min_guarantee_auto) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("Set a min_guarantee to disable "
                         "automatic min_guarantee"));
        goto cleanup;
    }

    memset(&mem_guarantee, 0, sizeof(mem_guarantee));
    if (set_min_guarantee) {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_BYTES;
        mem_guarantee.value = min_guarantee << 10; /* KiB to bytes */
    } else if (set_min_guarantee_auto) {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_AUTO;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    VZCT_CHECK(cleanup, vzctl2_env_set_memguarantee(param, &mem_guarantee));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (set_min_guarantee) {
        dom->def->mem.min_guarantee = min_guarantee;
        dom->def->mem.min_guarantee_auto = false;
        if (active) {
            dom->newDef->mem.min_guarantee = min_guarantee;
            dom->newDef->mem.min_guarantee_auto = false;
        }
    } else if (set_min_guarantee_auto) {
        dom->def->mem.min_guarantee = 0;
        dom->def->mem.min_guarantee_auto = true;
        if (active) {
            dom->newDef->mem.min_guarantee = 0;
            dom->newDef->mem.min_guarantee_auto = true;
        }
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (active) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetMemoryParameters(virDomainPtr domain,
                              virTypedParameterPtr params,
                              int *nparams,
                              unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get simultaneously persistent "
                         "and active configs"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetMemoryParametersEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get active config of inactive domain"));
        goto cleanup;
    }

    if (!*nparams) {
        *nparams = 1;
        ret = 0;
        goto cleanup;
    }

    if (!dom->def->mem.min_guarantee_auto) {
        if (virTypedParameterAssign(&params[0], VIR_DOMAIN_MEMORY_MIN_GUARANTEE,
                                    VIR_TYPED_PARAM_ULLONG,
                                    dom->def->mem.min_guarantee) < 0)
            goto cleanup;
    } else {
        if (virTypedParameterAssign(&params[0],
                                    VIR_DOMAIN_MEMORY_MIN_GUARANTEE_VZ_AUTO,
                                    VIR_TYPED_PARAM_BOOLEAN, true) < 0)
            goto cleanup;
    }

    *nparams = 1;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static void
vzctSetDefBlockIoTune(virDomainDefPtr def,
                      unsigned long long io,
                      unsigned long long iops,
                      bool set_io,
                      bool set_iops,
                      const char *group_name)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDefPtr disk = def->disks[i];
        virDomainBlockIoTuneInfoPtr iotune = &disk->blkdeviotune;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        if (set_io)
            iotune->total_bytes_sec = io;
        if (set_iops)
            iotune->total_iops_sec = iops;

        if (group_name) {
            VIR_FREE(iotune->group_name);
            iotune->group_name = g_strdup(group_name);
        } else if (def->ndisks > 1 && !iotune->group_name) {
            iotune->group_name = g_strdup("virtuozzo");
        }
    }
}


static int
vzctDomainSetBlockIoTune(virDomainPtr domain,
                         const char *disk,
                         virTypedParameterPtr params,
                         int nparams,
                         unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    virObjectEventPtr event;
    unsigned long long io, iops;
    bool set_io = false;
    bool set_iops = false;
    bool active = false;
    bool job = false;
    int rc, err, ret = -1;
    const char *group_name = NULL;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_BLOCK_IOTUNE_GROUP_NAME,
                               VIR_TYPED_PARAM_STRING,
                               NULL) < 0)
        return -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetBlockIoTuneEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        if (flags & (!(flags & VIR_DOMAIN_AFFECT_LIVE) ||
                     !(flags & VIR_DOMAIN_AFFECT_CONFIG))) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("Cannot modify independently persistent "
                             "and active configs"));
            goto cleanup;
        }

        active = true;
    } else {
        if (flags & VIR_DOMAIN_AFFECT_LIVE) {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("Cannot modify active config of inactive domain"));
            goto cleanup;
        }
    }

    if (!virDomainDiskByName(dom->def, disk, false)) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", disk);
        goto cleanup;
    }

    if ((rc = virTypedParamsGetULLong(params, nparams,
                                      VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                                      &io)) < 0)
        goto cleanup;
    if (rc == 1)
        set_io = true;

    if ((rc = virTypedParamsGetULLong(params, nparams,
                                      VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                                      &iops)) < 0)
        goto cleanup;
    if (rc == 1)
        set_iops = true;

    if (!set_io && !set_iops) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("You must set or change at least one limit"));
        goto cleanup;
    }

    if (virTypedParamsGetString(params, nparams,
                                VIR_DOMAIN_BLOCK_IOTUNE_GROUP_NAME,
                                &group_name) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                     0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    if (set_io)
        VZCT_CHECK(cleanup, vzctl2_env_set_iolimit(param, io));
    if (set_iops)
        VZCT_CHECK(cleanup, vzctl2_env_set_iopslimit(param, iops));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    vzctSetDefBlockIoTune(dom->def, io, iops, set_io, set_iops, group_name);

    if (active)
        vzctSetDefBlockIoTune(dom->newDef, io, iops, set_io, set_iops,
                              group_name);

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (active) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom, VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetBlockIoTune(virDomainPtr domain,
                         const char *disk,
                         virTypedParameterPtr params,
                         int *nparams,
                         unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    virDomainDiskDefPtr def;
    bool job = false;
    int param = 0;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get simultaneously persistent "
                         "and active configs"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetBlockIoTuneEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get active config of inactive domain"));
        goto cleanup;
    }

    if (!*nparams) {
        *nparams = 3;
        ret = 0;
        goto cleanup;
    }

    if (!(def = virDomainDiskByName(dom->def, disk, false))) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", disk);
        goto cleanup;
    }

    if (virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                                VIR_TYPED_PARAM_ULLONG,
                                def->blkdeviotune.total_bytes_sec) < 0)
        goto cleanup;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                                VIR_TYPED_PARAM_ULLONG,
                                def->blkdeviotune.total_iops_sec) < 0)
        goto cleanup;

    if (param < *nparams) {
        g_autofree char *group_name = NULL;

        group_name = g_strdup(def->blkdeviotune.group_name);
        if (virTypedParameterAssign(&params[param++],
                                    VIR_DOMAIN_BLOCK_IOTUNE_GROUP_NAME,
                                    VIR_TYPED_PARAM_STRING, group_name) < 0)
            goto cleanup;

        group_name = NULL;
    }

    *nparams = param;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainBlockResize(virDomainPtr domain,
                      const char *disk,
                      unsigned long long size,
                      unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    vzctl_env_param_ptr param;
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr dom = NULL;
    virDomainDiskDefPtr def;
    bool job = false;
    int layout, err, ret = -1;

    virCheckFlags(VIR_DOMAIN_BLOCK_RESIZE_BYTES, -1);

    if (flags & VIR_DOMAIN_BLOCK_RESIZE_BYTES)
        size = (size + 1023) / 1024; /* bytes to KiB */

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainBlockResizeEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!(def = virDomainDiskByName(dom->def, disk, false))) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", disk);
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    param = vzctl2_get_env_param(env);
    VZCT_CHECK(cleanup, vzctl2_env_get_layout(param, &layout));

    if (layout == VZCTL_LAYOUT_5) {
        struct vzctl_disk_param disk_param = { .size = size };
        const char *uuid = ((vzctDomainDiskPrivate *)def->privateData)->uuid;

        memcpy(disk_param.uuid, uuid, sizeof(disk_param.uuid));
        VZCT_CHECK(cleanup, vzctl2_env_set_disk(env, &disk_param));

        /* Also update diskspace for root disk image */
        if (STREQ(uuid, VZCT_DISK_ROOT_UUID)) {
            g_autoptr(vzctl_env_param) cparam = NULL;
            struct vzctl_2UL_res val = {size, size};

            VZCT_CHECK_PTR(cleanup, cparam, vzctl2_alloc_env_param());
            VZCT_CHECK(cleanup, vzctl2_env_set_diskspace(cparam, &val));
            VZCT_CHECK(cleanup, vzctl2_apply_param(env, cparam, VZCTL_SAVE));
        }
    } else {
        g_autoptr(vzctl_env_param) cparam = NULL;
        struct vzctl_2UL_res space = {size, size};
        struct vzctl_2UL_res inodes = {size / 4, size / 4};

        VZCT_CHECK_PTR(cleanup, cparam, vzctl2_alloc_env_param());
        VZCT_CHECK(cleanup, vzctl2_env_set_diskspace(cparam, &space));
        VZCT_CHECK(cleanup, vzctl2_env_set_diskinodes(cparam, &inodes));
        VZCT_CHECK(cleanup, vzctl2_apply_param(env, cparam, VZCTL_SAVE));
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetMetadata(virDomainPtr domain,
                      int type,
                      const char *metadata,
                      const char *key,
                      const char *uri,
                      unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainDefPtr persistentDef;
    virDomainDefPtr def;
    virObjectEventPtr event;
    virDomainObjPtr dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetMetadataEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (virDomainObjGetDefs(dom, flags, &def, &persistentDef) < 0)
        goto cleanup;

    if (def) {
        g_autofree char *statefile = vzctStateFile(driver->config, dom->def->uuid);

        if (virDomainDefSetMetadata(def, type, metadata, key, uri) < 0)
            goto cleanup;

        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    if (persistentDef) {
        if (virDomainDefSetMetadata(persistentDef, type, metadata, key, uri) < 0)
            goto cleanup;

        if (vzctDomainSaveConfig(driver, dom) < 0)
            goto cleanup;
    }

    event = virDomainEventMetadataChangeNewFromObj(dom, type, uri);
    virObjectEventStateQueue(driver->eventsManager, event);
    event = NULL;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);

    virDomainObjEndAPI(&dom);
    return ret;
}


static char *
vzctDomainGetMetadata(virDomainPtr domain,
                      int type,
                      const char *uri,
                      unsigned int flags)
{
    virDomainObjPtr dom;
    char *ret = NULL;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetMetadataEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = virDomainObjGetMetadata(dom, type, uri, flags);

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainRenameCallback(virDomainObjPtr dom,
                         const char *new_name,
                         unsigned int fflags G_GNUC_UNUSED,
                         void *opaque)
{
    vzctDriverPtr driver = opaque;
    g_autoptr(vzctl_env_handle) env = NULL;
    int err;

    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_RET(-1, vzctl2_set_name(env, new_name));

    VIR_FREE(dom->def->name);
    dom->def->name = g_strdup(new_name);

    if (virDomainObjIsActive(dom)) {
        VIR_FREE(dom->newDef->name);
        dom->newDef->name = g_strdup(new_name);
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        VIR_WARN("Cannot save config with new name: %s", new_name);

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            VIR_WARN("Cannot save status with new name: %s", new_name);
    }

    return 0;
}


static int
vzctDomainRename(virDomainPtr domain,
                 const char *new_name,
                 unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virObjectEventPtr event_new;
    virObjectEventPtr event_old;
    virDomainObjPtr dom = NULL;
    g_autofree char *old_name = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainRenameEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    old_name = g_strdup(dom->def->name);
    if (virDomainObjListRename(driver->domains, dom, new_name, flags,
                               vzctDomainRenameCallback, driver) < 0)
        goto cleanup;

    event_old = virDomainEventLifecycleNew(dom->def->id, old_name,
                                           dom->def->uuid,
                                           VIR_DOMAIN_EVENT_UNDEFINED,
                                           VIR_DOMAIN_EVENT_UNDEFINED_RENAMED);
    event_new = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_RENAMED);
    virObjectEventStateQueue(driver->eventsManager, event_old);
    virObjectEventStateQueue(driver->eventsManager, event_new);
    event_old = NULL;
    event_new = NULL;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static virHypervisorDriver vzctHypervisorDriver = {
    .domainCommandXExec = vzctDomainCommandXExec, /* 5.6.0 */
    .domainCommandXGetStatus = vzctDomainCommandXGetStatus, /* 5.6.0 */
    .domainCommandXTerminate = vzctDomainCommandXTerminate, /* 5.6.0 */
    .domainCommandXSetParams = vzctDomainCommandXSetParams, /* 5.6.0 */
    .name = "vzct",
    .connectOpen = vzctConnectOpen, /* 5.6.0 */
    .connectClose = vzctConnectClose, /* 5.6.0 */
    .connectSupportsFeature = vzctConnectSupportsFeature, /* 5.6.0 */
    .connectGetType = vzctConnectGetType,  /* 5.6.0 */
    .connectListAllDomains = vzctConnectListAllDomains, /* 5.6.0 */
    .connectGetCapabilities = vzctConnectGetCapabilities, /* 5.6.0 */
    .connectGetAllDomainStats = vzctConnectGetAllDomainStats, /* 5.6.0 */
    .connectDomainEventRegisterAny = vzctConnectDomainEventRegisterAny, /* 5.6.0 */
    .connectDomainEventDeregisterAny = vzctConnectDomainEventDeregisterAny, /* 5.6.0 */
    .domainGetState = vzctDomainGetState, /* 5.6.0 */
    .domainLookupByUUID = vzctDomainLookupByUUID, /* 5.6.0 */
    .domainLookupByName = vzctDomainLookupByName, /* 5.6.0 */
    .domainGetXMLDesc = vzctDomainGetXMLDesc, /* 5.6.0 */
    .domainDefineXML = vzctDomainDefineXML,      /* 5.6.0 */
    .domainDefineXMLFlags = vzctDomainDefineXMLFlags, /* 5.6.0 */
    .domainUndefine = vzctDomainUndefine, /* 5.6.0 */
    .domainUndefineFlags = vzctDomainUndefineFlags, /* 5.6.0 */
    .domainIsActive = vzctDomainIsActive, /* 5.6.0 */
    .domainCreate = vzctDomainCreate, /* 5.6.0 */
    .domainCreateWithFlags = vzctDomainCreateWithFlags, /* 5.6.0 */
    .domainDestroy = vzctDomainDestroy, /* 5.6.0 */
    .domainDestroyFlags = vzctDomainDestroyFlags, /* 5.6.0 */
    .domainShutdown = vzctDomainShutdown, /* 5.6.0 */
    .domainShutdownFlags = vzctDomainShutdownFlags, /* 5.6.0 */
    .domainGetInfo = vzctDomainGetInfo, /* 5.6.0 */
    .domainIsPersistent = vzctDomainIsPersistent, /* 5.6.0 */
    .domainGetOSType = vzctDomainGetOSType, /* 5.6.0 */
    .domainHasManagedSaveImage = vzctDomainHasManagedSaveImage, /* 5.6.0 */
    .domainSnapshotGetXMLDesc = vzctDomainSnapshotGetXMLDesc, /* 5.6.0 */
    .domainSnapshotNum = vzctDomainSnapshotNum, /* 5.6.0 */
    .domainSnapshotListNames = vzctDomainSnapshotListNames, /* 5.6.0 */
    .domainListAllSnapshots = vzctDomainListAllSnapshots, /* 5.6.0 */
    .domainSnapshotNumChildren = vzctDomainSnapshotNumChildren, /* 5.6.0 */
    .domainSnapshotListChildrenNames = vzctDomainSnapshotListChildrenNames, /* 5.6.0 */
    .domainSnapshotListAllChildren = vzctDomainSnapshotListAllChildren, /* 5.6.0 */
    .domainSnapshotLookupByName = vzctDomainSnapshotLookupByName, /* 5.6.0 */
    .domainHasCurrentSnapshot = vzctDomainHasCurrentSnapshot, /* 5.6.0 */
    .domainSnapshotGetParent = vzctDomainSnapshotGetParent, /* 5.6.0 */
    .domainSnapshotCurrent = vzctDomainSnapshotCurrent, /* 5.6.0 */
    .domainSnapshotIsCurrent = vzctDomainSnapshotIsCurrent, /* 5.6.0 */
    .domainSnapshotHasMetadata = vzctDomainSnapshotHasMetadata, /* 5.6.0 */
    .domainRevertToSnapshot = vzctDomainRevertToSnapshot, /* 5.6.0 */
    .domainSnapshotDelete = vzctDomainSnapshotDelete, /* 5.6.0 */
    .domainSnapshotCreateXML = vzctDomainSnapshotCreateXML, /* 5.6.0 */
    .domainMigrateBegin3Params = vzctDomainMigrateBegin3Params, /* 5.6.0 */
    .domainMigratePrepare3Params = vzctDomainMigratePrepare3Params, /* 5.6.0 */
    .domainMigratePerform3Params = vzctDomainMigratePerform3Params, /* 5.6.0 */
    .domainMigrateFinish3Params = vzctDomainMigrateFinish3Params, /* 5.6.0 */
    .domainMigrateConfirm3Params = vzctDomainMigrateConfirm3Params, /* 5.6.0 */
    .domainGetJobInfo = vzctDomainGetJobInfo, /* 5.6.0 */
    .domainGetJobStats = vzctDomainGetJobStats, /* 5.6.0 */
    .domainAbortJob = vzctDomainAbortJob, /* 5.6.0 */
    .domainGetAutostart = vzctDomainGetAutostart, /* 5.6.0 */
    .domainSetAutostart = vzctDomainSetAutostart, /* 5.6.0 */
    .domainSetMemoryFlags = vzctDomainSetMemoryFlags, /* 5.6.0 */
    .domainSetVcpusFlags = vzctDomainSetVcpusFlags, /* 5.6.0 */
    .domainGetVcpusFlags = vzctDomainGetVcpusFlags, /* 5.6.0 */
    .domainSetMemoryParameters = vzctDomainSetMemoryParameters, /* 5.6.0 */
    .domainGetMemoryParameters = vzctDomainGetMemoryParameters, /* 5.6.0 */
    .domainSetBlockIoTune = vzctDomainSetBlockIoTune, /* 5.6.0 */
    .domainGetBlockIoTune = vzctDomainGetBlockIoTune, /* 5.6.0 */
    .domainBlockResize = vzctDomainBlockResize, /* 5.6.0 */
    .domainSetMetadata = vzctDomainSetMetadata, /* 6.6.0 */
    .domainGetMetadata = vzctDomainGetMetadata, /* 6.6.0 */
    .domainRename = vzctDomainRename, /* 6.6.0 */
};

static virConnectDriver vzctConnectDriver = {
    .localOnly = true,
    .uriSchemes = (const char *[]){ "vzct", NULL },
    .hypervisorDriver = &vzctHypervisorDriver,
};

static virStateDriver vzctStateDriver = {
    .name = "vzct",
    .stateInitialize = vzctStateInitialize,
    .stateCleanup = vzctStateCleanup,
};


int
vzctRegister(void)
{
    if (virRegisterConnectDriver(&vzctConnectDriver, true) < 0)
        return -1;

    if (virRegisterStateDriver(&vzctStateDriver) < 0)
        return -1;

    return 0;
}
